import mmcv
import pytest

from otx.mpa.exporter_mixin import ExporterMixin
from tests.test_suite.e2e_test_system import e2e_pytest_unit


class TestExporterMixin:
    @pytest.fixture(autouse=True)
    def setup(self, mocker):
        def mock_init_logger():
            pass

        def mock_configure(model_cfg, model_ckpt, data_cfg, training=False, **kwargs):
            return mmcv.ConfigDict()

        self.exporter = ExporterMixin()
        self.exporter._init_logger = mock_init_logger
        self.exporter.configure = mock_configure
        self.exporter.mode = ["mock_mode", "train"]
        fake_config = mmcv.ConfigDict(work_dir="/path/work_dir", data=dict(test=dict(dataset=mocker.MagicMock())))
        mocker.patch.object(self.exporter, "configure", return_value=fake_config)
        mocker.patch("os.listdir")

    @e2e_pytest_unit
    def test_run_with_error_raise(self):
        return_value = self.exporter.run({}, "", {}, mode="mock_mode")

        assert "outputs" in return_value
        assert return_value["outputs"] is None
        assert "msg" in return_value

    @e2e_pytest_unit
    def test_run_without_deploy_cfg(self, mocker):
        def mock_naive_export(output_dir, model_builder, precision, cfg, model_name="model"):
            pass

        self.exporter.naive_export = mock_naive_export
        return_value = self.exporter.run({}, "", {}, mode="mock_mode", model_builder=mocker.MagicMock())

        assert "outputs" in return_value
        assert return_value["outputs"]["bin"] == "/path/work_dir/model.bin"
        assert return_value["outputs"]["xml"] == "/path/work_dir/model.xml"
        assert "msg" in return_value
        assert return_value["msg"] == ""

    @e2e_pytest_unit
    def test_run_with_deploy_cfg(self, mocker):
        def mock_mmdeploy_export(output_dir, model_builder, precision, cfg, deploy_cfg, model_name="model"):
            pass

        self.exporter.mmdeploy_export = mock_mmdeploy_export
        return_value = self.exporter.run(
            {}, "", {}, mode="mock_mode", model_builder=mocker.MagicMock(), deploy_cfg=mmcv.ConfigDict(deploy=True)
        )

        assert "outputs" in return_value
        assert return_value["outputs"]["bin"] == "/path/work_dir/model.bin"
        assert return_value["outputs"]["xml"] == "/path/work_dir/model.xml"
        assert "msg" in return_value
        assert return_value["msg"] == ""

    @e2e_pytest_unit
    def test_mmdeploy_export(self, mocker):
        from otx.mpa.deploy.apis import MMdeployExporter

        mock_export_openvino = mocker.patch.object(MMdeployExporter, "export2openvino")

        ExporterMixin.mmdeploy_export(
            "", None, "FP16", dict(), mmcv.ConfigDict(backend_config=dict(mo_options=dict(flags=[])))
        )

        mock_export_openvino.assert_called_once()



"""Collection of utils for data in Action Task."""

# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import os.path as osp
from collections import defaultdict
from typing import List, Optional

import numpy as np
from mmcv import ConfigDict

from otx.api.entities.annotation import (
    Annotation,
    AnnotationSceneEntity,
    AnnotationSceneKind,
)
from otx.api.entities.dataset_item import DatasetItemEntity
from otx.api.entities.id import ID
from otx.api.entities.label import Domain, LabelEntity
from otx.api.entities.scored_label import ScoredLabel
from otx.api.entities.shapes.rectangle import Rectangle
from otx.api.entities.subset import Subset
from otx.api.utils.argument_checks import check_input_parameters_type


@check_input_parameters_type()
def find_label_by_name(labels: List[LabelEntity], name: str, domain: Domain):
    """Return label from name."""
    matching_labels = [label for label in labels if label.name == name]
    if len(matching_labels) == 1:
        return matching_labels[0]
    if len(matching_labels) == 0:
        label = LabelEntity(name=name, domain=domain, id=ID(int(name)))
        labels.append(label)
        return label
    raise ValueError("Found multiple matching labels")


def load_cls_annotations(ann_file, data_root):
    """Load annotation file to get video information."""
    video_infos = []
    with open(ann_file, "r", encoding="UTF-8") as fin:
        for line in fin:
            line_split = line.strip().split()
            video_info = {}
            idx = 0
            # idx for frame_dir
            if line_split[0] == "#":
                continue
            frame_dir = line_split[idx]
            if data_root is not None:
                frame_dir = osp.join(data_root, frame_dir)
            video_info["frame_dir"] = frame_dir
            idx += 1
            # idx for total_frames
            # TODO Support offsets in dataset
            video_info["total_frames"] = int(line_split[idx])
            idx += 1
            # idx for label[s]
            # TODO Support multi-label setting
            label = [int(x) for x in line_split[idx:]]
            assert label, f"missing label in line: {line}"
            assert len(label) == 1
            video_info["label"] = label[0]
            video_infos.append(video_info)

    return video_infos


# pylint: disable=too-many-locals
def load_det_annotations(ann_file, data_root):
    """Load AVA annotations."""
    video_infos = []
    records_dict_by_img = defaultdict(list)
    with open(ann_file, "r", encoding="utf-8") as fin:
        for line in fin:
            line_split = line.strip().split(",")

            label = int(line_split[6])
            video_id = line_split[0]
            timestamp = int(line_split[1])
            img_key = f"{video_id},{timestamp}"

            entity_box = np.array(list(map(float, line_split[2:6])))
            entity_id = int(line_split[7])

            video_info = dict(
                video_id=video_id,
                timestamp=timestamp,
                entity_box=entity_box,
                label=label,
                entity_id=entity_id,
            )
            records_dict_by_img[img_key].append(video_info)

    for img_key in records_dict_by_img:
        video_id, timestamp = img_key.split(",")
        bboxes, labels, entity_ids = parse_img_record(records_dict_by_img[img_key])
        ann = dict(gt_bboxes=bboxes, gt_labels=labels, entity_ids=entity_ids)
        frame_dir = video_id
        if data_root is not None:
            frame_dir = osp.join(data_root, frame_dir)
        # FIXME Image shape is hard-coded, this will be replaced with CVAT format
        video_info = dict(
            frame_dir=frame_dir,
            video_id=video_id,
            timestamp=int(timestamp),
            img_key=img_key,
            ann=ann,
            width=320,
            height=240,
        )
        video_infos.append(video_info)

    return video_infos


def parse_img_record(img_records):
    """Accumulate and colligate bbox annotation info."""
    bboxes, labels, entity_ids = [], [], []
    while len(img_records) > 0:
        img_record = img_records[0]
        num_img_records = len(img_records)

        selected_records = [x for x in img_records if np.array_equal(x["entity_box"], img_record["entity_box"])]

        num_selected_records = len(selected_records)
        img_records = [x for x in img_records if not np.array_equal(x["entity_box"], img_record["entity_box"])]

        assert len(img_records) + num_selected_records == num_img_records

        bboxes.append(img_record["entity_box"])
        valid_labels = np.array([selected_record["label"] for selected_record in selected_records])
        labels.append(valid_labels)
        entity_ids.append(img_record["entity_id"])

    bboxes = np.stack(bboxes)
    entity_ids = np.stack(entity_ids)
    return bboxes, labels, entity_ids


# pylint: disable=too-many-locals
def load_cls_dataset(
    ann_file_path: str,
    data_root_dir: str,
    domain: Domain,
    subset: Subset = Subset.NONE,
    labels_list: Optional[List[LabelEntity]] = None,
):
    """Convert video annotation information into DatasetItemEntity."""
    dataset_items = []
    video_infos = load_cls_annotations(ann_file_path, data_root_dir)

    for video_info in video_infos:
        label = video_info.pop("label")
        label = find_label_by_name(labels_list, str(label), domain)
        shapes = [Annotation(Rectangle.generate_full_box(), [ScoredLabel(label)])]
        dataset_item = DatasetItemEntity(
            media=video_info,
            annotation_scene=AnnotationSceneEntity(annotations=shapes, kind=AnnotationSceneKind.ANNOTATION),
            subset=subset,
        )
        dataset_items.append(dataset_item)

    return dataset_items


# pylint: disable=too-many-locals
def load_det_dataset(
    ann_file_path: str,
    data_root_dir: str,
    domain: Domain,
    subset: Subset = Subset.NONE,
    labels_list: Optional[List[LabelEntity]] = None,
):
    """Convert video annotation information into DatasetItemEntity."""
    dataset_items = []
    video_infos = load_det_annotations(ann_file_path, data_root_dir)

    for video_info in video_infos:
        ann = video_info.pop("ann")
        # TODO Check use of entity_ids
        gt_bboxes = ann["gt_bboxes"]
        gt_labels = ann["gt_labels"]
        shapes = []
        for bbox, labels in zip(gt_bboxes, gt_labels):
            labels = [find_label_by_name(labels_list, str(label), domain) for label in labels]
            shapes.append(
                Annotation(
                    Rectangle(bbox[0], bbox[1], bbox[2], bbox[3]),
                    [ScoredLabel(label, probability=1.0) for label in labels],
                )
            )
        dataset_item = DatasetItemEntity(
            media=ConfigDict(video_info),
            annotation_scene=AnnotationSceneEntity(annotations=shapes, kind=AnnotationSceneKind.ANNOTATION),
            subset=subset,
        )
        dataset_items.append(dataset_item)

    return dataset_items



"""Initialization of streamer."""
# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from otx.api.usecases.exportable_code.streamer.streamer import (
    CameraStreamer,
    DirStreamer,
    ImageStreamer,
    InvalidInput,
    OpenError,
    ThreadedStreamer,
    VideoStreamer,
    get_streamer,
)

__all__ = [
    "CameraStreamer",
    "DirStreamer",
    "ImageStreamer",
    "ThreadedStreamer",
    "VideoStreamer",
    "InvalidInput",
    "OpenError",
    "get_streamer",
]



"""OTX Adapters - mmaction2."""

# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

from .data import OTXActionClsDataset, OTXActionDetDataset
from .models import register_action_backbones
from .utils import export_model, patch_config, set_data_classes

__all__ = ["OTXActionClsDataset", "OTXActionDetDataset", "patch_config", "set_data_classes", "export_model"]

register_action_backbones()



"""DetCon implementation for self-supervised learning.

Original papers:
- 'Efficient Visual Pretraining with Contrastive Detection', https://arxiv.org/abs/2103.10957
"""

# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#
# pylint: disable=unused-argument, invalid-name, unnecessary-pass, not-callable

from collections import OrderedDict
from typing import Any, Dict, List, Optional, Tuple, Union

import torch
import torch.distributed as dist
from mmcv.runner import load_checkpoint
from mmseg.models.builder import (  # pylint: disable=no-name-in-module
    SEGMENTORS,
    build_backbone,
    build_loss,
    build_neck,
)
from mmseg.ops import resize
from torch import nn

from otx.mpa.modules.models.segmentors.class_incr_encoder_decoder import (
    ClassIncrEncoderDecoder,
)
from otx.mpa.utils.logger import get_logger

logger = get_logger()


class MaskPooling(nn.Module):
    """Mask pooling module to filter each class with the same class.

    Args:
        num_classes (int): The number of classes to be considered as pseudo classes. Default: 256.
        num_samples (int): The number of samples to be sampled. Default: 16.
        downsample (int): The ratio of the mask size to the feature size. Default: 32.
        replacement (bool): Whether samples are drawn with replacement or not.
            It can be used when `num_classes` is small rather than `num_samples`. Default: True.
    """

    def __init__(
        self,
        num_classes: int = 256,
        num_samples: int = 16,
        downsample: int = 32,
        replacement: bool = True,
    ):

        super().__init__()

        self.num_classes = num_classes
        self.num_samples = num_samples
        self.replacement = replacement

        self.mask_ids = torch.arange(num_classes)
        self.pool = nn.AvgPool2d(kernel_size=downsample, stride=downsample)

    def pool_masks(self, masks: torch.Tensor):
        """Perform mask pooling and create binary masks.

        Args:
            masks (Tensor): Ground truth masks.

        Returns:
            Tensor: Pooled binary masks.
        """
        if masks.ndim < 4:
            masks = masks.unsqueeze(dim=1)

        masks = masks == self.mask_ids[None, :, None, None].to(masks.device)
        masks = self.pool(masks.to(torch.float))

        b, c, h, w = masks.shape
        masks = torch.reshape(masks, (b, c, h * w))
        masks = torch.argmax(masks, dim=1)
        masks = torch.eye(self.num_classes).to(masks.device)[masks]
        masks = torch.transpose(masks, 1, 2)

        return masks

    def sample_masks(self, masks: torch.Tensor):
        """Samples of which binary masks to use in the loss.

        Args:
            masks (Tensor): Pooled binary masks from `self.pool_masks`.

        Returns:
            tuple[Tensor, Tensor]: (sampled_masks, mask_ids),
                sampled binary masks and ids used to sample masks.
        """
        assert masks.ndim == 3

        batch_size = masks.shape[0]
        mask_exists = torch.greater(masks.sum(dim=-1), 1e-3)
        sel_masks = mask_exists.to(torch.float) + 1e-11

        mask_ids = torch.multinomial(sel_masks, num_samples=self.num_samples, replacement=self.replacement)
        sampled_masks = torch.stack([masks[b][mask_ids[b]] for b in range(batch_size)])

        return sampled_masks, mask_ids

    def forward(self, masks: torch.Tensor):
        """Forward function for mask pooling.

        Args:
            masks (Tensor): Ground truth masks to be sampled.

        Returns:
            tuple[Tensor, Tensor]: (sampled_masks, sampled_mask_ids),
                normalized sampled binary masks and ids used to sample masks.
        """
        binary_masks = self.pool_masks(masks)
        sampled_masks, sampled_mask_ids = self.sample_masks(binary_masks)
        areas = sampled_masks.sum(dim=-1, keepdim=True)
        sampled_masks = sampled_masks / torch.maximum(areas, torch.tensor(1.0, device=areas.device))

        return sampled_masks, sampled_mask_ids


# pylint: disable=too-many-arguments, dangerous-default-value, too-many-instance-attributes
@SEGMENTORS.register_module()
class DetConB(nn.Module):
    """Implementation of 'Efficient Visual Pretraining with Contrastive Detection' \
        (https://arxiv.org/abs/2103.10957).

    Args:
        backbone (dict): Config dict for module of backbone ConvNet.
        neck (dict, optional): Config dict for module of deep features to compact feature vectors.
            Default: None.
        head (dict, optional): Config dict for module of loss functions. Default: None.
        pretrained (str, optional): Path to pre-trained weights. Default: None.
        base_momentum (float): The base momentum coefficient for the target network.
            Default: 0.996.
        num_classes (int): The number of classes to be considered as pseudo classes. Default: 256.
        num_samples (int): The number of samples to be sampled. Default: 16.
        downsample (int): The ratio of the mask size to the feature size. Default: 32.
        input_transform (str): Input transform of features from backbone. Default: "resize_concat".
        in_index (list): Feature index to be used for DetCon if the backbone outputs
            multi-scale features wrapped by list or tuple. Default: [0].
        align_corners (bool): Whether apply `align_corners` during resize. Default: False.
        loss_cfg (dict): DetCon loss configuration.
    """

    def __init__(
        self,
        backbone: Dict[str, Any],
        neck: Optional[Dict[str, Any]] = None,
        head: Optional[Dict[str, Any]] = None,
        pretrained: Optional[str] = None,
        base_momentum: float = 0.996,
        num_classes: int = 256,
        num_samples: int = 16,
        downsample: int = 32,
        input_transform: str = "resize_concat",
        in_index: Union[List[int], int] = [0],
        align_corners: bool = False,
        loss_cfg: Optional[Dict[str, Any]] = None,
        **kwargs,
    ):
        super().__init__()

        self.base_momentum = base_momentum
        self.momentum = base_momentum
        self.num_classes = num_classes
        self.num_samples = num_samples
        self.downsample = downsample
        self.input_transform = input_transform
        self.in_index = in_index
        self.align_corners = align_corners

        # build backbone
        self.online_backbone = build_backbone(backbone)
        self.target_backbone = build_backbone(backbone)

        # build projector
        self.online_projector = build_neck(neck)
        self.target_projector = build_neck(neck)

        # build head with predictor
        self.predictor = build_neck(head)

        # set maskpooling
        self.mask_pool = MaskPooling(num_classes, num_samples, downsample)

        self.init_weights(pretrained=pretrained)

        # build detcon loss
        self.detcon_loss = build_loss(loss_cfg)

        # Hooks for super_type transparent weight save
        self._register_state_dict_hook(self.state_dict_hook)

    def init_weights(self, pretrained: Optional[str] = None):
        """Initialize the weights of model.

        Args:
            pretrained (str, optional): Path to pre-trained weights.
                Default: None.
        """
        if pretrained is not None:
            logger.info(f"load model from: {pretrained}")
            load_checkpoint(
                self.online_backbone,
                pretrained,
                strict=False,
                map_location=None,
                logger=logger,
                revise_keys=[(r"^backbone\.", "")],
            )

        # init backbone
        for param_ol, param_tgt in zip(self.online_backbone.parameters(), self.target_backbone.parameters()):
            param_tgt.data.copy_(param_ol.data)
            param_tgt.requires_grad = False
            param_ol.requires_grad = True

        # init projector
        self.online_projector.init_weights(init_linear="kaiming")
        for param_ol, param_tgt in zip(self.online_projector.parameters(), self.target_projector.parameters()):
            param_tgt.data.copy_(param_ol.data)
            param_tgt.requires_grad = False
            param_ol.requires_grad = True

        # init the predictor
        self.predictor.init_weights()

    @torch.no_grad()
    def _momentum_update(self):
        """Momentum update of the target network."""
        for param_ol, param_tgt in zip(self.online_backbone.parameters(), self.target_backbone.parameters()):
            param_tgt.data = param_tgt.data * self.momentum + param_ol.data * (1.0 - self.momentum)

        for param_ol, param_tgt in zip(self.online_projector.parameters(), self.target_projector.parameters()):
            param_tgt.data = param_tgt.data * self.momentum + param_ol.data * (1.0 - self.momentum)

    def transform_inputs(self, inputs: Union[List, Tuple]):
        """Transform inputs for decoder.

        Args:
            inputs (list, tuple): List (or tuple) of multi-level img features.

        Returns:
            Tensor: The transformed inputs.
        """
        # TODO (sungchul): consider tensor component, too
        if self.input_transform == "resize_concat" and isinstance(self.in_index, (list, tuple)):
            inputs = [inputs[i] for i in self.in_index]
            upsampled_inputs = [
                resize(input=x, size=inputs[0].shape[2:], mode="bilinear", align_corners=self.align_corners)
                for x in inputs
            ]
            inputs = torch.cat(upsampled_inputs, dim=1)
        elif self.input_transform == "multiple_select" and isinstance(self.in_index, (list, tuple)):
            inputs = [inputs[i] for i in self.in_index]
        else:
            if isinstance(self.in_index, (list, tuple)):
                self.in_index = self.in_index[0]
            inputs = inputs[self.in_index]  # type: ignore

        return inputs

    def extract_feat(self, img: torch.Tensor):
        """Extract features from images.

        Args:
            img (Tensor): Input image.

        Return:
            Tensor: Features from the online_backbone.
        """
        x = self.online_backbone(img)
        return x

    def sample_masked_feats(self, feats: Union[torch.Tensor, List, Tuple], masks: torch.Tensor, projector: nn.Module):
        """Sampled features from mask.

        Args:
            feats (list, tuple, Tensor): Features from the backbone.
            masks (Tensor): Ground truth masks to be sampled and to be used to filter `feats`.
            projector (nn.Module): Projector MLP.

        Returns:
            tuple[Tensor, Tensor]: (proj, sampled_mask_ids), features from the projector and ids used to sample masks.
        """
        if isinstance(feats, (list, tuple)) and len(feats) > 1:
            feats = self.transform_inputs(feats)

        # TODO (sungchul): consider self.input_transform == "multiple_select"
        sampled_masks, sampled_mask_ids = self.mask_pool(masks)

        b, c, h, w = feats.shape  # type: ignore
        feats = feats.reshape((b, c, h * w)).transpose(1, 2)  # type: ignore
        sampled_feats = sampled_masks @ feats
        sampled_feats = sampled_feats.reshape((-1, c))

        proj = projector(sampled_feats)

        return proj, sampled_mask_ids

    # pylint: disable=too-many-locals
    def forward_train(
        self, img: torch.Tensor, img_metas: List[Dict], gt_semantic_seg: torch.Tensor, return_embedding: bool = False
    ):
        """Forward function for training.

        Args:
            img (Tensor): Input images.
            img_metas (list[dict]): Input information.
            gt_semantic_seg (Tensor): Pseudo masks.
                It is used to organize features among the same classes.
            return_embedding (bool): Whether returning embeddings from the online backbone.
                It can be used for SupCon. Default: False.

        Returns:
            dict[str, Tensor]: A dictionary of loss components.
        """
        assert img.ndim == 5 and gt_semantic_seg.ndim == 5
        img1, img2 = img[:, 0], img[:, 1]
        mask1, mask2 = gt_semantic_seg[:, :, 0], gt_semantic_seg[:, :, 1]

        embd1 = self.online_backbone(img1)
        proj1, id1 = self.sample_masked_feats(embd1, mask1, self.online_projector)
        proj2, id2 = self.sample_masked_feats(self.online_backbone(img2), mask2, self.online_projector)

        with torch.no_grad():
            self._momentum_update()
            proj1_tgt, id1_tgt = self.sample_masked_feats(self.target_backbone(img1), mask1, self.target_projector)
            proj2_tgt, id2_tgt = self.sample_masked_feats(self.target_backbone(img2), mask2, self.target_projector)

        # predictor
        # TODO (sungchul): predictor + loss -> head?
        pred1, pred2 = self.predictor(proj1), self.predictor(proj2)
        pred1 = pred1.reshape((-1, self.num_samples, pred1.shape[-1]))
        pred2 = pred2.reshape((-1, self.num_samples, pred2.shape[-1]))
        proj1_tgt = proj1_tgt.reshape((-1, self.num_samples, proj1_tgt.shape[-1]))
        proj2_tgt = proj2_tgt.reshape((-1, self.num_samples, proj2_tgt.shape[-1]))

        # decon loss
        loss = self.detcon_loss(
            pred1=pred1,
            pred2=pred2,
            target1=proj1_tgt,
            target2=proj2_tgt,
            pind1=id1,
            pind2=id2,
            tind1=id1_tgt,
            tind2=id2_tgt,
        )

        if return_embedding:
            return loss, embd1
        return loss

    def forward(self, img, img_metas, return_loss=True, **kwargs):
        """Calls either :func:`forward_train` or :func:`forward_test` depending on whether ``return_loss`` is ``True``.

        Note this setting will change the expected inputs. When
        ``return_loss=True``, img and img_meta are single-nested (i.e. Tensor
        and List[dict]), and when ``resturn_loss=False``, img and img_meta
        should be double nested (i.e.  List[Tensor], List[List[dict]]), with
        the outer list indicating test time augmentations.
        """
        if return_loss:
            return self.forward_train(img, img_metas, **kwargs)
        raise AttributeError("Self-SL doesn't support `forward_test` for evaluation.")

    def train_step(self, data_batch: Dict[str, Any], optimizer: Union[torch.optim.Optimizer, Dict], **kwargs):
        """The iteration step during training.

        This method defines an iteration step during training, except for the
        back propagation and optimizer updating, which are done in an optimizer
        hook. Note that in some complicated cases or models, the whole process
        including back propagation and optimizer updating is also defined in
        this method, such as GAN.

        Args:
            data_batch (dict): The output of dataloader.
            optimizer (:obj:`torch.optim.Optimizer` | dict): The optimizer of
                runner is passed to ``train_step()``. This argument is unused
                and reserved.

        Returns:
            dict: It should contain at least 3 keys: ``loss``, ``log_vars``,
                ``num_samples``.
                ``loss`` is a tensor for back propagation, which can be a
                weighted sum of multiple losses.
                ``log_vars`` contains all the variables to be sent to the
                logger.
                ``num_samples`` indicates the batch size (when the model is
                DDP, it means the batch size on each GPU), which is used for
                averaging the logs.
        """
        losses = self(**data_batch)
        loss, log_vars = self._parse_losses(losses)

        outputs = dict(loss=loss, log_vars=log_vars, num_samples=len(data_batch["img_metas"]))

        return outputs

    def val_step(self, **kwargs):
        """Disenable validation step during self-supervised learning."""
        pass

    def _parse_losses(self, losses: Dict[str, Any]):
        """Parse the raw outputs (losses) of the network.

        Args:
            losses (dict): Raw output of the network, which usually contain
                losses and other necessary information.

        Returns:
            tuple[Tensor, dict]: (loss, log_vars), loss is the loss tensor
                which may be a weighted sum of all losses, log_vars contains
                all the variables to be sent to the logger.
        """
        log_vars = OrderedDict()
        for var_name, var_value in losses.items():
            if isinstance(var_value, torch.Tensor):
                log_vars[var_name] = var_value.mean()
            elif isinstance(var_value, list):
                log_vars[var_name] = sum(_loss.mean() for _loss in var_value)
            elif isinstance(var_value, (int, float)):
                log_vars[var_name] = var_value
            else:
                raise TypeError(f"{var_name} is not a tensor or list of tensors")

        loss = sum(_value for _key, _value in log_vars.items() if "loss" in _key)

        log_vars["loss"] = loss
        for var_name, var_value in log_vars.items():
            if isinstance(var_value, (int, float)):
                continue

            # reduce loss when distributed training
            if dist.is_available() and dist.is_initialized():
                var_value = var_value.data.clone()
                dist.all_reduce(var_value.div_(dist.get_world_size()))

            log_vars[var_name] = var_value.item()

        return loss, log_vars

    def set_step_params(self, init_iter, epoch_size):
        """`set_step_params` to be skipped."""
        pass

    @staticmethod
    def state_dict_hook(module, state_dict, *args, **kwargs):
        """Save only online backbone as output state_dict."""
        logger.info("----------------- BYOL.state_dict_hook() called")
        output = OrderedDict()
        for k, v in state_dict.items():
            if "online_backbone." in k:
                k = k.replace("online_backbone.", "backbone.")
                output[k] = v
        return output


# pylint: disable=too-many-locals
@SEGMENTORS.register_module()
class SupConDetConB(ClassIncrEncoderDecoder):  # pylint: disable=too-many-ancestors
    """Apply DetConB as a contrastive part of `Supervised Contrastive Learning` (https://arxiv.org/abs/2004.11362).

    SupCon with DetConB uses ground truth masks instead of pseudo masks to organize features among the same classes.

    Args:
        decode_head (dict, optional): Config dict for module of decode head. Default: None.
        train_cfg (dict, optional): Config dict for training. Default: None.
    """

    def __init__(
        self,
        backbone: Dict[str, Any],
        decode_head: Optional[Dict[str, Any]] = None,
        neck: Optional[Dict[str, Any]] = None,
        head: Optional[Dict[str, Any]] = None,
        pretrained: Optional[str] = None,
        base_momentum: float = 0.996,
        num_classes: int = 256,
        num_samples: int = 16,
        downsample: int = 32,
        input_transform: str = "resize_concat",
        in_index: Union[List[int], int] = [0],
        align_corners: bool = False,
        loss_cfg: Optional[Dict[str, Any]] = None,
        train_cfg: Optional[Dict[str, Any]] = None,
        test_cfg: Optional[Dict[str, Any]] = None,
        **kwargs,
    ):
        super().__init__(backbone=backbone, decode_head=decode_head, train_cfg=train_cfg, test_cfg=test_cfg, **kwargs)

        self.detconb = DetConB(
            backbone=backbone,
            neck=neck,
            head=head,
            pretrained=pretrained,
            base_momentum=base_momentum,
            num_classes=num_classes,
            num_samples=num_samples,
            downsample=downsample,
            input_transform=input_transform,
            in_index=in_index,
            align_corners=align_corners,
            loss_cfg=loss_cfg,
            **kwargs,
        )
        self.backbone = self.detconb.online_backbone
        # TODO (sungchul): Is state_dict_hook needed to save segmentor only?
        # 1. use state_dict_hook : we can save memory as only saving backbone + decode_head.
        # 2. save all : we can use additional training with the whole weights (backbone + decode_head + detcon).

    # pylint: disable=arguments-renamed
    def forward_train(
        self,
        img: torch.Tensor,
        img_metas: List[Dict],
        gt_semantic_seg: torch.Tensor,
        pixel_weights: Optional[torch.Tensor] = None,
        **kwargs,
    ):
        """Forward function for training.

        Args:
            img (Tensor): Input images.
            img_metas (list[dict]): Input information.
            gt_semantic_seg (Tensor): Ground truth masks.
                It is used to organize features among the same classes.
            pixel_weights (Tensor): Pixels weights.

        Returns:
            dict[str, Tensor]: A dictionary of loss components.
        """
        losses = {}
        if img.ndim == 4:
            # supervised learning with interval
            embd = self.detconb.online_backbone(img)
            mask = gt_semantic_seg
        else:
            # supcon training
            mask = gt_semantic_seg[:, :, 0]
            loss_detcon, embd = self.detconb.forward_train(
                img=img, img_metas=img_metas, gt_semantic_seg=gt_semantic_seg, return_embedding=True
            )
            losses.update(dict(loss_detcon=loss_detcon["loss"]))

        # decode head
        loss_decode, _ = self._decode_head_forward_train(
            embd, img_metas, gt_semantic_seg=mask, pixel_weights=pixel_weights
        )
        losses.update(loss_decode)

        return losses



"""This module implements the OptimizationParameters entity."""
# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from dataclasses import dataclass
from typing import Any, Callable, Optional


# pylint: disable=unused-argument
def default_progress_callback(progress: int, score: Optional[float] = None):
    """This is the default progress callback for OptimizationParameters."""


def default_save_model_callback():
    """This is the default save model callback for OptimizationParameters."""


@dataclass
class OptimizationParameters:
    """Optimization parameters.

    resume (bool): Set to ``True`` if optimization must be resume with the optimizer state;
        set to ``False`` to discard the optimizer state and start with fresh optimizer
    update_progress (Callable[int], None): Callback which can be used to provide updates about the progress of a task.
    save_model (Callable[List, None]): Callback to notify that the model weights have been changed.
        This callback can be used by the task when temporary weights should be saved (for instance, at the
        end of an epoch). If this callback has been used to save temporary weights, those weights will be
        used to resume optimization if for some reason training was suspended.
    """

    resume: bool = False
    update_progress: Callable[[int, Optional[float]], Any] = default_progress_callback
    save_model: Callable[[], None] = default_save_model_callback



"""Collection of utils for task implementation in Segmentation Task."""

# Copyright (C) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.


import logging
import math
from typing import List, Optional, Union

from mmcv import Config, ConfigDict

from otx.algorithms.common.adapters.mmcv.utils import (
    get_configs_by_keys,
    get_configs_by_pairs,
    get_dataset_configs,
    get_meta_keys,
    is_epoch_based_runner,
    patch_color_conversion,
    prepare_work_dir,
    remove_from_config,
    remove_from_configs_by_type,
    update_config,
)
from otx.algorithms.segmentation.configs.base import SegmentationConfig
from otx.api.entities.label import Domain, LabelEntity
from otx.api.utils.argument_checks import (
    DirectoryPathCheck,
    check_input_parameters_type,
)

logger = logging.getLogger(__name__)


@check_input_parameters_type({"work_dir": DirectoryPathCheck})
def patch_config(
    config: Config,
    work_dir: str,
    labels: List[LabelEntity],
):  # pylint: disable=too-many-branches
    """Update config function."""

    # Add training cancelation hook.
    if "custom_hooks" not in config:
        config.custom_hooks = []
    if "CancelTrainingHook" not in {hook.type for hook in config.custom_hooks}:
        config.custom_hooks.append(ConfigDict({"type": "CancelTrainingHook"}))

    # Remove high level data pipelines definition leaving them only inside `data` section.
    remove_from_config(config, "train_pipeline")
    remove_from_config(config, "test_pipeline")
    # Remove cancel interface hook
    remove_from_configs_by_type(config.custom_hooks, "CancelInterfaceHook")

    config.evaluation["efficient_test"] = True
    evaluation_metric = config.evaluation.get("metric")
    if evaluation_metric is not None:
        config.evaluation.save_best = evaluation_metric

    config.checkpoint_config.max_keep_ckpts = 5
    config.checkpoint_config.interval = config.evaluation.get("interval", 1)

    label_names = ["background"] + [label.name for label in labels]
    set_data_classes(config, label_names)

    remove_from_config(config, "img_norm_cfg")

    config.gpu_ids = range(1)
    config.work_dir = work_dir


@check_input_parameters_type()
def patch_model_config(
    config: Config,
    labels: List[LabelEntity],
    distributed: bool = False,
):
    """Patch model config."""
    label_names = ["background"] + [label.name for label in labels]
    set_num_classes(config, len(label_names))

    if "test_cfg" not in config.model:
        config.model.test_cfg = ConfigDict()
    config.model.test_cfg.return_repr_vector = True
    set_distributed_mode(config, distributed)


@check_input_parameters_type()
def set_hyperparams(config: Config, hyperparams: SegmentationConfig):
    """Set function for hyperparams (SegmentationConfig)."""
    config.data.samples_per_gpu = int(hyperparams.learning_parameters.batch_size)
    config.data.workers_per_gpu = int(hyperparams.learning_parameters.num_workers)
    config.optimizer.lr = float(hyperparams.learning_parameters.learning_rate)

    # set proper number of iterations
    fixed_iters = int(hyperparams.learning_parameters.learning_rate_fixed_iters)
    warmup_iters = int(hyperparams.learning_parameters.learning_rate_warmup_iters)
    main_iters = int(hyperparams.learning_parameters.num_iters)
    total_iterations = fixed_iters + warmup_iters + main_iters

    # false positive (mypy)
    if config.lr_config.get("policy", None) == "customstep":
        config.lr_config.fixed_iters = fixed_iters
    config.find_unused_parameters = fixed_iters > 0
    config.lr_config.warmup_iters = warmup_iters
    if config.lr_config.warmup_iters == 0:
        config.lr_config.warmup = None
    if is_epoch_based_runner(config.runner):
        init_num_iterations = config.runner.max_epochs
        config.runner.max_epochs = total_iterations
    else:
        init_num_iterations = config.runner.max_iters
        config.runner.max_iters = total_iterations

    # rescale the learning schedules
    schedule_scale = float(total_iterations) / float(init_num_iterations) if init_num_iterations > 0 else 0.0
    rescale_num_iterations(config, schedule_scale)


@check_input_parameters_type()
def rescale_num_iterations(config: Union[Config, ConfigDict], schedule_scale: float):
    """Rescale number of iterations for lr scheduler."""
    if config.lr_config.policy == "customstep":
        config.lr_config.step = [int(schedule_scale * step) for step in config.lr_config.step]
    elif config.lr_config.policy == "customcos":
        config.lr_config.periods = [int(schedule_scale * period) for period in config.lr_config.periods]

    def _rescale_num_iters(head, schedule_scale):
        losses = head.loss_decode
        if not isinstance(losses, (tuple, list)):
            losses = [losses]

        for loss in losses:
            for target_attr in ["scale_cfg", "conf_penalty_weight"]:
                if not hasattr(loss, target_attr):
                    continue

                if not hasattr(loss[target_attr], "num_iters"):
                    continue

                loss[target_attr].num_iters = int(schedule_scale * loss[target_attr].num_iters)
        head.loss_decode = losses

    # rescale number of iterations for
    for head_type in ["decode_head", "auxiliary_head"]:
        if not hasattr(config.model, head_type):
            continue

        heads = config.model[head_type]
        if isinstance(heads, (tuple, list)):
            for head in heads:
                _rescale_num_iters(head, schedule_scale)
        elif isinstance(heads, dict):
            _rescale_num_iters(heads, schedule_scale)
        config.model[head_type] = heads


@check_input_parameters_type()
def patch_adaptive_repeat_dataset(
    config: Union[Config, ConfigDict], num_samples: int, decay: float = 0.002, factor: float = 10
):
    """Patch the repeat times and training epochs adatively."""
    if config.data.train.type != "RepeatDataset":
        return

    if not config.data.train.get("adaptive_repeat", False):
        return

    if not is_epoch_based_runner(config.runner):
        return

    # calculate new number of iterations
    init_max_epoch = config.runner.max_epochs
    new_repeat_times = max(round(math.exp(-decay * num_samples) * factor), 1)
    new_max_epoch = math.ceil(init_max_epoch / new_repeat_times)
    if new_max_epoch <= 1:
        return

    # set proper number of iterations
    config.runner.max_epochs = new_max_epoch
    config.data.train.times = new_repeat_times

    # rescale the learning schedules
    schedule_scale = float(new_max_epoch) / float(init_max_epoch)
    config.params_config.iters = math.ceil(schedule_scale * config.params_config.iters)
    config.lr_config.fixed_iters = math.ceil(schedule_scale * config.lr_config.fixed_iters)
    config.lr_config.warmup_iters = math.ceil(schedule_scale * config.lr_config.warmup_iters)
    rescale_num_iterations(config, schedule_scale)


@check_input_parameters_type()
def prepare_for_training(
    config: Config,
    data_config: ConfigDict,
) -> Config:
    """Prepare configs for training phase."""
    prepare_work_dir(config)

    train_num_samples = 0
    for subset in ["train", "val", "test"]:
        data_config_ = data_config.data.get(subset)
        config_ = config.data.get(subset)
        if data_config_ is None:
            continue
        for key in ["otx_dataset", "labels"]:
            found = get_configs_by_keys(data_config_, key, return_path=True)
            if len(found) == 0:
                continue
            assert len(found) == 1
            if subset == "train" and key == "otx_dataset":
                found_value = list(found.values())[0]
                if found_value:
                    train_num_samples = len(found_value)
            update_config(config_, found)

    if train_num_samples > 0:
        patch_adaptive_repeat_dataset(config, train_num_samples)

    return config


@check_input_parameters_type()
def set_distributed_mode(config: Config, distributed: bool):
    """Setter distributed into config."""
    if distributed:
        return

    norm_cfg = {"type": "BN", "requires_grad": True}

    def _replace_syncbn(_node, _norm_cfg):
        if _node.norm_cfg.type != "SyncBN":
            return

        _node.norm_cfg = _norm_cfg

    config.model.backbone.norm_cfg = norm_cfg
    for head_type in ("decode_head", "auxiliary_head"):
        head = config.model.get(head_type, None)
        if head is None:
            continue

        if isinstance(head, (tuple, list)):
            for sub_head in head:
                _replace_syncbn(sub_head, norm_cfg)
        else:
            _replace_syncbn(head, norm_cfg)


@check_input_parameters_type()
def set_data_classes(config: Config, label_names: List[str]):
    """Setter data classes into config."""
    # Save labels in data configs.
    for subset in ("train", "val", "test"):
        cfg = config.data[subset]
        if cfg.type == "RepeatDataset":
            cfg.dataset.classes = label_names
        else:
            cfg.classes = label_names
        config.data[subset].classes = label_names


@check_input_parameters_type()
def set_num_classes(config: Config, num_classes: int):
    """Setter num_classes function."""
    assert num_classes > 1

    for head_type in ("decode_head", "auxiliary_head"):
        heads = config.model.get(head_type, None)
        if heads is None:
            continue

        if isinstance(heads, (tuple, list)):
            for head in heads:
                if hasattr(head, "loss_target") and head.loss_target == "gt_class_borders":
                    continue

                head.num_classes = num_classes
        elif isinstance(heads, dict):
            heads["num_classes"] = num_classes


@check_input_parameters_type()
def patch_datasets(
    config: Config,
    domain: Domain = Domain.SEGMENTATION,
    subsets: Optional[List[str]] = None,
    **kwargs,
):
    """Update dataset configs."""
    assert "data" in config
    assert "type" in kwargs

    if subsets is None:
        subsets = ["train", "val", "test", "unlabeled"]

    def update_pipeline(cfg):
        if subset == "train":
            for collect_cfg in get_configs_by_pairs(cfg, dict(type="Collect")):
                get_meta_keys(collect_cfg)
        for cfg_ in get_configs_by_pairs(cfg, dict(type="LoadImageFromFile")):
            cfg_.type = "LoadImageFromOTXDataset"
        for cfg_ in get_configs_by_pairs(cfg, dict(type="LoadAnnotations")):
            cfg_.type = "LoadAnnotationFromOTXDataset"

    for subset in subsets:
        if subset not in config.data:
            continue
        config.data[f"{subset}_dataloader"] = config.data.get(f"{subset}_dataloader", ConfigDict())

        cfgs = get_dataset_configs(config, subset)
        for cfg in cfgs:
            cfg.domain = domain
            cfg.otx_dataset = None
            cfg.labels = None
            cfg.update(kwargs)

            remove_from_config(cfg, "ann_dir")
            remove_from_config(cfg, "img_dir")
            remove_from_config(cfg, "data_root")
            remove_from_config(cfg, "split")
            remove_from_config(cfg, "classes")

            update_pipeline(cfg)

    patch_color_conversion(config)


def patch_evaluation(config: Config):
    """Update evaluation configs."""
    cfg = config.get("evaluation", None)
    if cfg:
        cfg.pop("classwise", None)
        cfg.metric = "mDice"
        cfg.save_best = "mDice"
        cfg.rule = "greater"
        # EarlyStoppingHook
        config.early_stop_metric = "mDice"



"""NNCF task related hooks."""
# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import math

from mmcv.runner.hooks import HOOKS, LrUpdaterHook


# pylint: disable=abstract-method,too-many-instance-attributes
class BaseLrUpdaterHook(LrUpdaterHook):
    """BaseLrUpdaterHook."""

    schedulers = ["constant", "semi-constant", "linear", "cos"]

    def __init__(  # pylint: disable=too-many-arguments
        self,
        by_epoch=True,
        fixed=None,
        fixed_iters=0,
        fixed_ratio=1.0,
        warmup=None,
        warmup_iters=0,
        warmup_ratio=0.1,
    ):
        super().__init__(by_epoch, warmup, warmup_iters, warmup_ratio)

        if fixed is not None:
            assert fixed in self.schedulers
            assert fixed_iters >= 0

        if warmup is not None:
            assert warmup in self.schedulers
            assert warmup_iters >= 0
            assert 0 < warmup_ratio <= 1.0

        self.warmup_iters = warmup_iters if warmup else 0

        self.fixed_policy = fixed
        self.fixed_iters = fixed_iters if fixed else 0
        self.fixed_start_ratio = fixed_ratio
        self.fixed_end_ratio = self.warmup_ratio if warmup is not None else 1.0

        self.base_lr = []
        self.epoch_len = None
        self.need_update = True

    @staticmethod
    def _get_lr(policy, cur_iters, regular_lr, max_iters, start_scale, end_scale):
        progress = float(cur_iters) / float(max_iters)
        if policy == "constant":
            k = start_scale
        elif policy == "semi-constant":
            threshold = 0.8
            if progress < threshold:
                k = start_scale
            else:
                progress = (progress - threshold) / (1.0 - threshold)
                k = (end_scale - start_scale) * progress + start_scale
        elif policy == "linear":
            k = (end_scale - start_scale) * progress + start_scale
        elif policy == "cos":
            k = end_scale + 0.5 * (start_scale - end_scale) * (math.cos(math.pi * progress) + 1.0)
        else:
            raise ValueError(f"Unknown policy: {policy}")

        return [_lr * k for _lr in regular_lr]

    def get_regular_lr(self, runner):
        """get_regular_lr."""
        if isinstance(runner.optimizer, dict):
            lr_groups = {}
            for k in runner.optimizer.keys():
                _lr_group = [self.get_lr(runner, _base_lr) for _base_lr in self.base_lr[k]]
                lr_groups.update({k: _lr_group})

            return lr_groups
        return [self.get_lr(runner, _base_lr) for _base_lr in self.base_lr]

    def get_fixed_lr(self, cur_iters, regular_lr):
        """get_fixed_lr."""
        return self._get_lr(
            self.fixed_policy, cur_iters, regular_lr, self.fixed_iters, self.fixed_start_ratio, self.fixed_end_ratio
        )

    def get_warmup_lr(self, cur_iters, regular_lr):
        """get_warmup_lr."""
        return self._get_lr(self.warmup, cur_iters, regular_lr, self.warmup_iters, self.warmup_ratio, 1.0)

    def _init_states(self, runner):
        if self.by_epoch:
            self.epoch_len = len(runner.data_loader)
            assert self.epoch_len > 0

            self.fixed_iters = self.fixed_iters * self.epoch_len
            self.warmup_iters = self.warmup_iters * self.epoch_len
        else:
            self.epoch_len = 1

        runner.model.module.set_step_params(runner.iter, self.epoch_len)

    def before_train_iter(self, runner):
        """before_train_iter."""
        if self.need_update:
            self._init_states(runner)
            self.need_update = False

        cur_iter = runner.iter
        regular_lr = self.get_regular_lr(runner)

        if cur_iter >= self.warmup_iters + self.fixed_iters:
            self._set_lr(runner, regular_lr)
        elif cur_iter >= self.fixed_iters:
            warmup_lr = self.get_warmup_lr(cur_iter - self.fixed_iters, regular_lr)
            self._set_lr(runner, warmup_lr)
        else:
            fixed_lr = self.get_fixed_lr(cur_iter, regular_lr)
            self._set_lr(runner, fixed_lr)


@HOOKS.register_module()
class CustomstepLrUpdaterHook(BaseLrUpdaterHook):
    """CustomstepLrUpdaterHook."""

    def __init__(self, step, gamma=0.1, **kwargs):
        super().__init__(**kwargs)

        assert isinstance(step, (list, int))
        if isinstance(step, list):
            for s in step:  # pylint: disable=invalid-name
                assert isinstance(s, int) and s >= 0
        elif isinstance(step, int):
            assert step >= 0
        else:
            raise TypeError('"step" must be a list or integer')

        self.steps = step if isinstance(step, (tuple, list)) else [step]
        self.gamma = gamma

    def _init_states(self, runner):
        super()._init_states(runner)

        if self.by_epoch:
            self.steps = [step * self.epoch_len for step in self.steps]

    def get_lr(self, runner, base_lr):
        """get_lr."""
        progress = runner.iter

        skip_iters = self.fixed_iters + self.warmup_iters
        if progress <= skip_iters:
            return base_lr

        exp = len(self.steps)
        for i, step in enumerate(self.steps):
            if progress < step:
                exp = i
                break

        return base_lr * self.gamma**exp



# Copyright (C) 2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#


"""
The functions in the file generate pytest decorators
for integrating with e2e test system and the class DataCollector
that allows pushing information to the dashboard of e2e test system.

If e2e test system is not installed, the generated pytest decorators do nothing,
and the DataCollector class is replaced with a stub that does nothing too.
"""

import functools
import traceback

import pytest


def _generate_e2e_pytest_decorators():
    try:
        from e2e.markers.mark_meta import MarkMeta
    except ImportError:

        def _e2e_pytest_api(func):
            return func

        def _e2e_pytest_performance(func):
            return func

        def _e2e_pytest_component(func):
            return func

        def _e2e_pytest_unit(func):
            return func

        return (
            _e2e_pytest_api,
            _e2e_pytest_performance,
            _e2e_pytest_component,
            _e2e_pytest_unit,
        )

    class Requirements:
        # Dummy requirement
        REQ_DUMMY = "Dummy requirement"

    class OTXComponent(MarkMeta):
        OTX = "otx"

    def _e2e_pytest_api(func):
        @pytest.mark.components(OTXComponent.OTX)
        @pytest.mark.priority_medium
        @pytest.mark.reqids(Requirements.REQ_DUMMY)
        @pytest.mark.api_other
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            return func(*args, **kwargs)

        return wrapper

    def _e2e_pytest_performance(func):
        @pytest.mark.components(OTXComponent.OTX)
        @pytest.mark.priority_medium
        @pytest.mark.reqids(Requirements.REQ_DUMMY)
        @pytest.mark.api_performance
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            return func(*args, **kwargs)

        return wrapper

    def _e2e_pytest_component(func):
        @pytest.mark.components(OTXComponent.OTX)
        @pytest.mark.priority_medium
        @pytest.mark.reqids(Requirements.REQ_DUMMY)
        @pytest.mark.component
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            return func(*args, **kwargs)

        return wrapper

    def _e2e_pytest_unit(func):
        @pytest.mark.components(OTXComponent.OTX)
        @pytest.mark.priority_medium
        @pytest.mark.reqids(Requirements.REQ_DUMMY)
        @pytest.mark.unit
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            return func(*args, **kwargs)

        return wrapper

    return (
        _e2e_pytest_api,
        _e2e_pytest_performance,
        _e2e_pytest_component,
        _e2e_pytest_unit,
    )


def _create_class_DataCollector():
    try:
        from e2e.collection_system.systems import TinySystem

        return TinySystem
    except ImportError:

        class _dummy_DataCollector:  # should have the same interface as TinySystem
            def __init__(self, *args, **kwargs):
                pass

            def flush(self):
                pass

            def register_collector(self, *args, **kwargs):
                pass

            def register_exporter(self, *args, **kwargs):
                pass

            def log_final_metric(self, *args, **kwargs):
                pass

            def log_internal_metric(self, *args, **kwargs):
                pass

            def update_metadata(self, *args, **kwargs):
                pass

            def __enter__(self):
                return self

            def __exit__(self, typerr, value, tback):
                if typerr is not None:
                    traceback.format_tb(tback)
                    raise typerr(value)
                return True

        return _dummy_DataCollector


(
    e2e_pytest_api,
    e2e_pytest_performance,
    e2e_pytest_component,
    e2e_pytest_unit,
) = _generate_e2e_pytest_decorators()
DataCollector = _create_class_DataCollector()



"""ModelContainer class used for loading the model in the model wrapper."""
# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import importlib
from pathlib import Path
from typing import Any, Tuple

import numpy as np
from openvino.model_zoo.model_api.adapters import OpenvinoAdapter, create_core
from openvino.model_zoo.model_api.models import Model

from otx.api.entities.label_schema import LabelSchemaEntity
from otx.api.entities.model_template import TaskType
from otx.api.serialization.label_mapper import LabelSchemaMapper
from otx.api.utils import Tiler
from otx.api.utils.detection_utils import detection2array

from .utils import get_model_path, get_parameters


class ModelContainer:
    """Class for storing the model wrapper based on Model API and needed parameters of model.

    Args:
        model_dir (Path): path to model directory
    """

    def __init__(self, model_dir: Path, device="CPU") -> None:
        self.parameters = get_parameters(model_dir / "config.json")
        self._labels = LabelSchemaMapper.backward(self.parameters["model_parameters"]["labels"])
        self._task_type = TaskType[self.parameters["converter_type"]]

        self.segm = bool(
            self._task_type is TaskType.ROTATED_DETECTION or self._task_type is TaskType.INSTANCE_SEGMENTATION
        )

        # labels for modelAPI wrappers can be empty, because unused in pre- and postprocessing
        self.model_parameters = self.parameters["model_parameters"]
        self.model_parameters["labels"] = []

        model_adapter = OpenvinoAdapter(create_core(), get_model_path(model_dir / "model.xml"), device=device)

        self._initialize_wrapper()
        self.core_model = Model.create_model(
            self.parameters["type_of_model"],
            model_adapter,
            self.model_parameters,
            preload=True,
        )

        self.tiler = self.setup_tiler()

    def setup_tiler(self):
        """Setup tiler.

        Returns:
            Tiler: tiler module
        """
        if (
            not self.parameters.get("tiling_parameters")
            or not self.parameters["tiling_parameters"]["enable_tiling"]["value"]
        ):
            return None

        tile_size = self.parameters["tiling_parameters"]["tile_size"]["value"]
        tile_overlap = self.parameters["tiling_parameters"]["tile_overlap"]["value"]
        max_number = self.parameters["tiling_parameters"]["tile_max_number"]["value"]
        tiler = Tiler(tile_size, tile_overlap, max_number, self.core_model, self.segm)
        return tiler

    @property
    def task_type(self) -> TaskType:
        """Task type property."""
        return self._task_type

    @property
    def labels(self) -> LabelSchemaEntity:
        """Labels property."""
        return self._labels

    @staticmethod
    def _initialize_wrapper() -> None:
        """Load the model class."""
        try:
            importlib.import_module("model_wrappers")
        except ModuleNotFoundError:
            print("Using model wrapper from Open Model Zoo ModelAPI")

    def infer(self, frame):
        """Infer with original image.

        Args:
            frame (np.ndarray): image
        Returns:
            annotation_scene (AnnotationScene): prediction
            frame_meta (Dict): dict with original shape
        """
        # getting result include preprocessing, infer, postprocessing for sync infer
        predictions, frame_meta = self.core_model(frame)

        # MaskRCNN returns tuple so no need to process
        if self._task_type == TaskType.DETECTION:
            predictions = detection2array(predictions)
        return predictions, frame_meta

    def infer_tile(self, frame):
        """Infer by patching full image to tiles.

        Args:
            frame (np.ndarray): image
        Returns:
            annotation_scene (AnnotationScene): prediction
            frame_meta (Dict): dict with original shape
        """

        detections, _ = self.tiler.predict(frame)
        return detections, {"original_shape": frame.shape}

    def __call__(self, input_data: np.ndarray) -> Tuple[Any, dict]:
        """Infer entry wrapper."""
        if self.tiler:
            return self.infer_tile(input_data)
        return self.infer(input_data)



"""This module contains the ConfigurableEnum, that is used to define Enums which can be configured by the user."""
# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from enum import Enum
from typing import List

from .metadata_keys import ENUM_NAME, OPTIONS


class ConfigurableEnum(Enum):
    """This class is used as the basis for defining `selectable` configurable parameters in the OTX API.

    Enums reflecting `selectable` options should inherit from thisclass.
    """

    def __str__(self):
        """Retrieves the string representation of an instance of the ConfigurableEnum (or subclasses thereof)."""
        return self.value

    def __eq__(self, other) -> bool:
        """Returns True if the ConfigurableEnum instance is equal to the other ConfigurableEnum instance.

        Checks whether one ConfigurableEnum instance (or instance of a subclass thereof) is equal to the `other`
        object. Comparison is made based on class name, instance value and instance name.

        Returns:
            bool: True if the two instances are equal, False otherwise.
        """
        if isinstance(other, ConfigurableEnum) and self.__class__.__name__ == other.__class__.__name__:
            return self.value == other.value and self.name == other.name
        return False

    def __hash__(self):
        """Computes hash for the ConfigurableEnum instance."""
        return hash(self.name)

    @classmethod
    def get_class_info(cls) -> dict:
        """Creates a dictionary representation of the ConfigurableEnum.

        This includes the name of the enum and the (name, value) pairs representing its members.

        Returns:
            dict: Dictionary representation of the ConfigurableEnum.
        """
        options_dict = {name: instance.value for name, instance in cls.__members__.items()}

        return {ENUM_NAME: cls.__name__, OPTIONS: options_dict}

    @classmethod
    def get_names(cls) -> List[str]:
        """Returns a list of names that can be used to index the Enum."""
        return [x.name for x in cls]

    @classmethod
    def get_values(cls) -> List[str]:
        """Returns a list of values that can be used to index the Enum."""
        return [x.value for x in cls]



"""Patch mmseg library."""
# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

# false positive pylint: disable=no-name-in-module
from mmseg.models.segmentors.base import BaseSegmentor

from otx.algorithms.common.adapters.nncf.patches import nncf_trace_context

# add nncf context method that will be used when nncf tracing
BaseSegmentor.nncf_trace_context = nncf_trace_context



"""
Test configurable parameters for the anomaly classification task
"""

# Copyright (C) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import logging
import os

import pytest

from otx.algorithms.anomaly.adapters.anomalib.config import get_anomalib_config
from otx.algorithms.anomaly.configs.padim import PadimAnomalyBaseConfig
from otx.algorithms.anomaly.configs.stfpm import STFPMAnomalyBaseConfig
from otx.api.configuration.helper import convert, create
from tests.helpers.config import get_config_and_task_name

logger = logging.getLogger(__name__)


@pytest.mark.parametrize(
    ["model_name", "configurable_parameters"],
    [("padim", PadimAnomalyBaseConfig), ("stfpm", STFPMAnomalyBaseConfig)],
)
def test_configuration_yaml(configurable_parameters, model_name):
    # assert that we can parse the template.yaml
    template_file_path = os.path.join("configs", "anomaly_classification", model_name, "template.yaml")
    configuration_yaml_loaded, task_name = get_config_and_task_name(template_file_path)

    configuration = configurable_parameters()
    # assert that we can convert our config object to yaml format
    configuration_yaml_str = convert(configuration, str)
    # assert that we can create configurable parameters from the yaml string
    configuration_yaml_converted = create(configuration_yaml_str)
    # assert that we generate an anomalib config from the
    get_anomalib_config(task_name, configuration_yaml_converted)
    # assert that the python class and the yaml file result in the same configurable parameters object
    assert configuration_yaml_converted == configuration_yaml_loaded



import numpy as np
import pytest
import torch
from mmcv.parallel import DataContainer

from otx.mpa.modules.datasets.pipelines.transforms.seg_custom_pipelines import (
    BranchImage,
    DefaultFormatBundle,
    Normalize,
)
from tests.test_suite.e2e_test_system import e2e_pytest_unit


class TestNormalize:
    @e2e_pytest_unit
    @pytest.mark.parametrize(
        "mean,std,to_rgb,expected",
        [
            (1.0, 1.0, True, np.array([[[1.0, 0.0, 0.0]]], dtype=np.float32)),
            (1.0, 1.0, False, np.array([[[-1.0, 0.0, 0.0]]], dtype=np.float32)),
        ],
    )
    def test_call(self, mean: float, std: float, to_rgb: bool, expected: np.array) -> None:
        """Test __call__."""
        normalize = Normalize(mean=mean, std=std, to_rgb=to_rgb)
        inputs = dict(img=np.arange(3).reshape(1, 1, 3))

        results = normalize(inputs.copy())

        assert "img" in results
        assert "img_norm_cfg" in results
        assert np.all(results["img"] == expected)

    @e2e_pytest_unit
    @pytest.mark.parametrize("mean,std,to_rgb", [(1.0, 1.0, True)])
    def test_repr(self, mean: float, std: float, to_rgb: bool) -> None:
        """Test __repr__."""
        normalize = Normalize(mean=mean, std=std, to_rgb=to_rgb)

        assert repr(normalize) == normalize.__class__.__name__ + f"(mean={mean}, std={std}, to_rgb=" f"{to_rgb})"


class TestDefaultFormatBundle:
    @pytest.fixture(autouse=True)
    def setup(self) -> None:
        self.default_format_bundle = DefaultFormatBundle()

    @e2e_pytest_unit
    @pytest.mark.parametrize("img", [np.ones((1, 1)), np.ones((1, 1, 1)), np.ones((1, 1, 1, 1))])
    @pytest.mark.parametrize("gt_semantic_seg,pixel_weights", [(np.ones((1, 1)), np.ones((1, 1)))])
    def test_call(self, img: np.array, gt_semantic_seg: np.array, pixel_weights: np.array) -> None:
        """Test __call__."""
        inputs = dict(img=img, gt_semantic_seg=gt_semantic_seg, pixel_weights=pixel_weights)

        results = self.default_format_bundle(inputs.copy())

        assert isinstance(results, dict)
        assert "img" in results
        assert isinstance(results["img"], DataContainer)
        assert len(results["img"].data.shape) >= 3
        assert results["img"].data.dtype == torch.float32
        assert "gt_semantic_seg" in results
        assert len(results["gt_semantic_seg"].data.shape) == len(inputs["gt_semantic_seg"].shape) + 1
        assert results["gt_semantic_seg"].data.dtype == torch.int64
        assert "pixel_weights" in results
        assert len(results["pixel_weights"].data.shape) == len(inputs["pixel_weights"].shape) + 1
        assert results["pixel_weights"].data.dtype == torch.float32

    @e2e_pytest_unit
    @pytest.mark.parametrize("img", [np.ones((1,))])
    def test_call_invalid_shape(self, img: np.array):
        inputs = dict(img=img)

        with pytest.raises(ValueError):
            self.default_format_bundle(inputs.copy())

    @e2e_pytest_unit
    def test_repr(self) -> None:
        """Test __repr__."""
        assert repr(self.default_format_bundle) == self.default_format_bundle.__class__.__name__


class TestBranchImage:
    @pytest.fixture(autouse=True)
    def setup(self) -> None:
        self.branch_image = BranchImage(key_map={"key1": "key2"})

    @e2e_pytest_unit
    def test_call(self) -> None:
        """Test __call__."""
        inputs = dict(key1="key1", img_fields=["key1"])

        results = self.branch_image(inputs.copy())

        assert isinstance(results, dict)
        assert "key2" in results
        assert results["key1"] == results["key2"]
        assert "key2" in results["img_fields"]

    @e2e_pytest_unit
    def test_repr(self) -> None:
        """Test __repr__."""
        assert repr(self.branch_image) == self.branch_image.__class__.__name__



# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import copy

import numpy as np
import pytest
from openvino.model_zoo.model_api.models import Model

import otx.algorithms.classification.tasks.openvino
from otx.algorithms.classification.configs.base import ClassificationConfig
from otx.algorithms.classification.tasks.openvino import (
    ClassificationOpenVINOInferencer,
    ClassificationOpenVINOTask,
)
from otx.api.configuration.configurable_parameters import ConfigurableParameters
from otx.api.entities.annotation import (
    Annotation,
    AnnotationSceneEntity,
    AnnotationSceneKind,
)
from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.label_schema import LabelSchemaEntity
from otx.api.entities.metrics import Performance, ScoreMetric
from otx.api.entities.model import ModelConfiguration, ModelEntity
from otx.api.entities.resultset import ResultSetEntity
from otx.api.entities.scored_label import ScoredLabel
from otx.api.entities.shapes.rectangle import Rectangle
from otx.api.usecases.evaluation.metrics_helper import MetricsHelper
from otx.api.usecases.tasks.interfaces.optimization_interface import OptimizationType
from otx.api.utils.shape_factory import ShapeFactory
from tests.test_suite.e2e_test_system import e2e_pytest_unit
from tests.unit.algorithms.classification.test_helper import (
    DEFAULT_CLS_TEMPLATE,
    init_environment,
    setup_configurable_parameters,
)


@pytest.fixture
def otx_model():
    model_configuration = ModelConfiguration(
        configurable_parameters=ConfigurableParameters(header="header", description="description"),
        label_schema=LabelSchemaEntity(),
    )
    return ModelEntity(train_dataset=DatasetEntity(), configuration=model_configuration)


class TestOpenVINOClassificationInferencer:
    @pytest.fixture(autouse=True)
    def setup(self, mocker) -> None:
        hyper_parameters, model_template = setup_configurable_parameters(DEFAULT_CLS_TEMPLATE)
        cls_params = ClassificationConfig(header=hyper_parameters.header)
        environment, dataset = init_environment(hyper_parameters, model_template)
        self.label_schema = environment.label_schema
        mocker.patch("otx.algorithms.classification.tasks.openvino.OpenvinoAdapter")
        mocker.patch.object(Model, "create_model")
        self.cls_ov_inferencer = ClassificationOpenVINOInferencer(cls_params, self.label_schema, "")
        model_path = "otx.algorithms.classification.adapters.openvino.model_wrappers.openvino_models.OTXClassification"
        self.cls_ov_inferencer.model = mocker.patch(model_path, autospec=True)
        self.fake_input = np.random.rand(3, 224, 224)

    @e2e_pytest_unit
    def test_pre_process(self):
        self.cls_ov_inferencer.model.preprocess.return_value = {"foo": "bar"}
        returned_value = self.cls_ov_inferencer.pre_process(self.fake_input)

        assert returned_value == {"foo": "bar"}

    @e2e_pytest_unit
    def test_post_process(self):
        fake_feature_vector = np.random.rand(1, 100)
        fake_logits = np.random.rand(1, 2)
        fake_prediction = {
            "logits": fake_logits,
            "feature_vector": fake_feature_vector,
            "saliency_map": self.fake_input,
        }
        fake_metadata = {"original_shape": (254, 320, 3), "resized_shape": (224, 224, 3)}
        self.cls_ov_inferencer.model.postprocess.return_value = [[0, 0.87], [1, 0.13]]
        returned_value = self.cls_ov_inferencer.post_process(fake_prediction, fake_metadata)

        assert len(returned_value.annotations[0].get_labels()) > 0
        assert len(returned_value.annotations[0].get_labels()) == len(self.label_schema.get_labels(include_empty=False))

    @e2e_pytest_unit
    def test_predict(self, mocker):
        fake_output = AnnotationSceneEntity(kind=AnnotationSceneKind.ANNOTATION, annotations=[])
        mock_pre_process = mocker.patch.object(ClassificationOpenVINOInferencer, "pre_process", return_value=("", ""))
        mock_forward = mocker.patch.object(ClassificationOpenVINOInferencer, "forward")
        self.cls_ov_inferencer.model.postprocess_aux_outputs.return_value = ("", "", "", "")
        mock_post_process = mocker.patch.object(
            ClassificationOpenVINOInferencer, "post_process", return_value=fake_output
        )
        returned_value = self.cls_ov_inferencer.predict(self.fake_input)

        mock_pre_process.assert_called_once()
        mock_forward.assert_called_once()
        mock_post_process.assert_called_once()
        assert returned_value == (fake_output, "", "", "", "")

    @e2e_pytest_unit
    def test_forward(self):
        fake_output = {"logits": np.random.rand(1, 2)}
        self.cls_ov_inferencer.model.infer_sync.return_value = fake_output
        returned_value = self.cls_ov_inferencer.forward({"image": self.fake_input})
        assert returned_value == fake_output


class TestOpenVINOClassificationTask:
    @pytest.fixture(autouse=True)
    def setup(self, mocker, otx_model) -> None:
        hyper_parameters, model_template = setup_configurable_parameters(DEFAULT_CLS_TEMPLATE)
        cls_params = ClassificationConfig(header=hyper_parameters.header)
        self.task_env, self.dataset = init_environment(params=hyper_parameters, model_template=model_template)
        self.label_schema = self.task_env.label_schema
        mocker.patch("otx.algorithms.classification.tasks.openvino.OpenvinoAdapter")
        mocker.patch.object(Model, "create_model")
        cls_ov_inferencer = ClassificationOpenVINOInferencer(cls_params, self.label_schema, "")
        self.task_env.model = otx_model
        mocker.patch.object(ClassificationOpenVINOTask, "load_inferencer", return_value=cls_ov_inferencer)
        self.cls_ov_task = ClassificationOpenVINOTask(self.task_env)
        self.labels = self.label_schema.get_labels(include_empty=True)
        fake_annotation = [
            Annotation(
                Rectangle.generate_full_box(),
                id=0,
                labels=[ScoredLabel(label, probability=1.0) for label in self.labels],
            )
        ]
        self.fake_ann_scene = AnnotationSceneEntity(kind=AnnotationSceneKind.ANNOTATION, annotations=fake_annotation)
        self.fake_input = mocker.MagicMock()

    @e2e_pytest_unit
    def test_infer(self, mocker):
        mock_predict = mocker.patch.object(
            ClassificationOpenVINOInferencer,
            "predict",
            return_value=(self.fake_ann_scene, np.array([0, 1]), self.fake_input, self.fake_input, self.fake_input),
        )
        mocker.patch.object(ShapeFactory, "shape_produces_valid_crop", return_value=True)
        updated_dataset = self.cls_ov_task.infer(self.dataset)

        mock_predict.assert_called()
        for updated in updated_dataset:
            assert updated.annotation_scene.contains_any(self.labels)

    @e2e_pytest_unit
    def test_explain(self, mocker):
        self.fake_silency_map = np.random.randint(255, size=(2, 224, 224), dtype=np.uint8)
        mocker.patch.object(
            ClassificationOpenVINOInferencer,
            "predict",
            return_value=(
                self.fake_ann_scene,
                np.array([0, 1]),
                self.fake_silency_map,
                self.fake_input,
                self.fake_input,
            ),
        )
        updpated_dataset = self.cls_ov_task.explain(self.dataset)

        assert updpated_dataset is not None
        assert updpated_dataset.get_labels() == self.dataset.get_labels()

    @e2e_pytest_unit
    def test_evaluate(self, mocker):
        result_set = ResultSetEntity(
            model=None,
            ground_truth_dataset=DatasetEntity(),
            prediction_dataset=DatasetEntity(),
        )
        fake_metrics = mocker.patch("otx.api.usecases.evaluation.accuracy.Accuracy", autospec=True)
        fake_metrics.get_performance.return_value = Performance(
            score=ScoreMetric(name="fake", value=0.1), dashboard_metrics="Accuracy"
        )
        mocker.patch.object(MetricsHelper, "compute_accuracy", return_value=fake_metrics)
        self.cls_ov_task.evaluate(result_set)

        assert result_set.performance.score.value == 0.1

    @e2e_pytest_unit
    def test_deploy(self, otx_model):
        output_model = copy.deepcopy(otx_model)
        self.cls_ov_task.model.set_data("openvino.bin", b"foo")
        self.cls_ov_task.model.set_data("openvino.xml", b"bar")
        self.cls_ov_task.deploy(output_model)

        assert output_model.exportable_code is not None

    @e2e_pytest_unit
    def test_optimize(self, mocker, otx_model):
        def patch_save_model(model, dir_path, model_name):
            with open(f"{dir_path}/{model_name}.xml", "wb") as f:
                f.write(b"foo")
            with open(f"{dir_path}/{model_name}.bin", "wb") as f:
                f.write(b"bar")

        output_model = copy.deepcopy(otx_model)
        self.cls_ov_task.model.set_data("openvino.bin", b"foo")
        self.cls_ov_task.model.set_data("openvino.xml", b"bar")
        mocker.patch("otx.algorithms.classification.tasks.openvino.load_model", autospec=True)
        mocker.patch("otx.algorithms.classification.tasks.openvino.create_pipeline", autospec=True)
        mocker.patch("otx.algorithms.classification.tasks.openvino.save_model", new=patch_save_model)
        spy_compress = mocker.spy(otx.algorithms.classification.tasks.openvino, "compress_model_weights")
        self.cls_ov_task.optimize(OptimizationType.POT, dataset=self.dataset, output_model=output_model)

        spy_compress.assert_called_once()
        assert self.cls_ov_task.model.get_data("openvino.bin")
        assert self.cls_ov_task.model.get_data("openvino.xml")



"""Unit Tests for the MPA Dataset Pipelines - Random Augment."""

# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


import numpy as np
import pytest
from PIL import Image

from otx.mpa.modules.datasets.pipelines.transforms.random_augment import (
    CutoutAbs,
    MPARandAugment,
    rand_augment_pool,
)


@pytest.fixture
def sample_np_image() -> np.ndarray:
    return np.ones((256, 256, 3), dtype=np.uint8)


@pytest.fixture
def sample_pil_image() -> Image:
    return Image.new("RGB", (256, 256), (255, 255, 255))


def test_all_transforms_return_valid_image(sample_pil_image: Image.Image) -> None:
    """Test all transforms return valid image."""
    for transform, v, max_v in rand_augment_pool:
        img, *extra = transform(sample_pil_image, v=v, max_v=max_v)
        assert isinstance(img, Image.Image)
        assert img.size == sample_pil_image.size


def test_cutoutabs_transform(sample_pil_image: Image.Image) -> None:
    """Test CutoutAbs transform."""
    img, (x0, y0, x1, y1), color = CutoutAbs(sample_pil_image, 2)
    assert isinstance(img, Image.Image)
    assert img.size == sample_pil_image.size
    assert x0 >= 0 and y0 >= 0
    assert x1 <= sample_pil_image.width and y1 <= sample_pil_image.height
    assert color == (127, 127, 127)


class TestMPARandAugment:
    def test_with_default_arguments(self, sample_np_image: np.ndarray) -> None:
        """Test case with default arguments."""
        transform = MPARandAugment(n=2, m=5, cutout=16)
        data = {"img": sample_np_image}
        results = transform(data)

        assert isinstance(results["img"], np.ndarray)
        assert any(item.startswith("rand_mc_") for item in results.keys())
        assert "CutoutAbs" in results

    def test_with_img_fields_argument(self, sample_np_image: np.ndarray) -> None:
        """Test case with img_fields argument."""
        transform = MPARandAugment(n=2, m=5, cutout=16)
        data = {
            "img1": sample_np_image,
            "img2": sample_np_image,
            "img_fields": ["img1"],
        }
        results = transform(data)
        assert isinstance(results["img1"], np.ndarray)
        assert any(item.startswith("rand_mc_") for item in results.keys())
        assert "CutoutAbs" in results

    def test_with_pil_image_input(self, sample_pil_image: Image.Image) -> None:
        """Test case with PIL.Image input."""
        transform = MPARandAugment(n=2, m=5, cutout=16)
        data = {"img": sample_pil_image}
        results = transform(data)

        assert isinstance(results["img"], np.ndarray)
        assert "CutoutAbs" in results



# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from typing import Dict

import pytest
import torch
import torch.nn as nn

from otx.algorithms.segmentation.adapters.mmseg import SelfSLMLP
from tests.test_suite.e2e_test_system import e2e_pytest_unit


class TestSelfSLMLP:
    @e2e_pytest_unit
    @pytest.mark.parametrize("use_conv", [True, False])
    @pytest.mark.parametrize("with_avg_pool", [True, False])
    def test_init(self, use_conv: bool, with_avg_pool: bool):
        """Test __init__ function."""
        selfslmlp = SelfSLMLP(
            in_channels=2, hid_channels=2, out_channels=2, use_conv=use_conv, with_avg_pool=with_avg_pool
        )

        if with_avg_pool:
            assert isinstance(selfslmlp.avgpool, nn.AdaptiveAvgPool2d)

        if use_conv:
            assert isinstance(selfslmlp.mlp[0], nn.Conv2d)
            assert isinstance(selfslmlp.mlp[3], nn.Conv2d)
        else:
            assert isinstance(selfslmlp.mlp[0], nn.Linear)
            assert isinstance(selfslmlp.mlp[3], nn.Linear)

    @e2e_pytest_unit
    @pytest.mark.parametrize("init_linear", ["normal", "kaiming"])
    @pytest.mark.parametrize("std", [0.01, 1.0, 10.0])
    @pytest.mark.parametrize("bias", [0.1, 0.0, 1.0])
    def test_init_weights(self, init_linear: str, std: float, bias: float):
        """Test init_weights function.

        Check if weights of nn.Linear was changed except for biases.
        BatchNorm weights are already set to 1, so it isn't required to be checked.
        """

        def gather_weight_mean_std(modules):
            mean = []
            std = []
            for module in modules:
                if isinstance(module, nn.Linear):
                    mean.append(module.weight.mean())
                    std.append(module.weight.std())
            return mean, std

        selfslmlp = SelfSLMLP(in_channels=2, hid_channels=2, out_channels=2, use_conv=False, with_avg_pool=True)

        orig_mean, orig_std = gather_weight_mean_std(selfslmlp.modules())

        selfslmlp.init_weights(init_linear, std, bias)

        updated_mean, updated_std = gather_weight_mean_std(selfslmlp.modules())

        for origs, updateds in zip([orig_mean, orig_std], [updated_mean, updated_std]):
            for orig, updated in zip(origs, updateds):
                assert orig != updated

    @e2e_pytest_unit
    @pytest.mark.parametrize("init_linear", ["undefined"])
    def test_init_weights_undefined_initialization(self, init_linear: str):
        """Test init_weights function when undefined initialization is given."""
        selfslmlp = SelfSLMLP(in_channels=2, hid_channels=2, out_channels=2, use_conv=False, with_avg_pool=True)

        with pytest.raises(ValueError):
            selfslmlp.init_weights(init_linear)

    @e2e_pytest_unit
    @pytest.mark.parametrize(
        "inputs,norm_cfg,use_conv,with_avg_pool,expected",
        [
            (torch.rand((2, 2)), dict(type="BN1d"), False, False, torch.Size([2, 2])),
            (torch.rand((2, 2, 2, 2)), dict(type="BN1d"), False, True, torch.Size([2, 2])),
            (torch.rand((2, 2, 2, 2)), dict(type="BN2d"), True, False, torch.Size([2, 2, 2, 2])),
            (torch.rand((2, 2, 2, 2)), dict(type="BN2d"), True, True, torch.Size([2, 2, 1, 1])),
        ],
    )
    def test_forward_tensor(
        self, inputs: torch.Tensor, norm_cfg: Dict, use_conv: bool, with_avg_pool: bool, expected: torch.Size
    ):
        """Test forward function for tensor."""
        selfslmlp = SelfSLMLP(
            in_channels=2,
            hid_channels=2,
            out_channels=2,
            norm_cfg=norm_cfg,
            use_conv=use_conv,
            with_avg_pool=with_avg_pool,
        )

        results = selfslmlp(inputs)

        assert results.shape == expected

    @e2e_pytest_unit
    @pytest.mark.parametrize(
        "inputs,norm_cfg,use_conv,with_avg_pool,expected",
        [
            ([torch.rand((2, 2)), torch.rand((2, 2))], dict(type="BN1d"), False, False, torch.Size([2, 2])),
            ([torch.rand((2, 2, 2, 2)), torch.rand((2, 2, 2, 2))], dict(type="BN1d"), False, True, torch.Size([2, 2])),
            (
                [torch.rand((2, 2, 2, 2)), torch.rand((2, 2, 2, 2))],
                dict(type="BN2d"),
                True,
                False,
                torch.Size([2, 2, 2, 2]),
            ),
            (
                [torch.rand((2, 2, 2, 2)), torch.rand((2, 2, 2, 2))],
                dict(type="BN2d"),
                True,
                True,
                torch.Size([2, 2, 1, 1]),
            ),
        ],
    )
    def test_forward_list_tuple(
        self, inputs: torch.Tensor, norm_cfg: Dict, use_conv: bool, with_avg_pool: bool, expected: torch.Size
    ):
        """Test forward function for list or tuple."""
        selfslmlp = SelfSLMLP(
            in_channels=2,
            hid_channels=2,
            out_channels=2,
            norm_cfg=norm_cfg,
            use_conv=use_conv,
            with_avg_pool=with_avg_pool,
        )

        results = selfslmlp(inputs)

        assert results.shape == expected

    @e2e_pytest_unit
    @pytest.mark.parametrize("inputs", ["unsupported", 1])
    def test_forward_unsupported_format(self, inputs: str):
        """Test forward function for unsupported format."""
        selfslmlp = SelfSLMLP(
            in_channels=2,
            hid_channels=2,
            out_channels=2,
        )

        with pytest.raises(TypeError):
            selfslmlp(inputs)



"""Patch mmcls."""
# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from mmcls.models.classifiers.base import BaseClassifier

from otx.algorithms.common.adapters.nncf.patches import nncf_trace_context

# add nncf context method that will be used when nncf tracing
BaseClassifier.nncf_trace_context = nncf_trace_context



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import errno
import os
from typing import List, Optional

from openvino.pyopenvino import Model, Node
from openvino.runtime import Core

from otx.mpa.utils.logger import get_logger

from .omz_wrapper import AVAILABLE_OMZ_MODELS, get_omz_model

logger = get_logger()


def to_dynamic_model(ov_model: Model) -> Model:
    assert isinstance(ov_model, Model)

    shapes = {}
    target_layouts = {}
    for input_node in ov_model.inputs:
        target_layout = {
            "batch": ["N", None, None],
            "height": ["H", None, None],
            "width": ["W", None, None],
        }

        any_name = input_node.any_name
        parameter_node = input_node.get_node()
        layout = parameter_node.get_layout()
        if layout.empty:
            continue
        layout = layout.to_string()[1:-1].split(",")
        shape = [str(i) for i in input_node.get_partial_shape()]
        for i, (layout_name, shape_) in enumerate(zip(layout, shape)):
            try:
                shape_ = int(shape_)
            except ValueError:
                shape_ = -1

            for target_layout_ in target_layout.values():
                target_layout_name = target_layout_[0]
                if layout_name == target_layout_name:
                    target_layout_[1] = i
                    target_layout_[2] = shape_
                    shape_ = -1
                    break
            shape[i] = shape_

        shapes[any_name] = shape
        target_layouts[any_name] = target_layout

    def reshape_model(ov_model, shapes):
        try:
            ov_model.reshape(shapes)
            return True
        except Exception:
            return False

    pop_targets = [["height", "width"], ["batch"]]
    pop_targets = pop_targets[::-1]
    while not reshape_model(ov_model, shapes):
        for key in shapes.keys():
            shape = shapes[key]
            target_layout = target_layouts[key]

            targets = pop_targets.pop()
            for target in targets:
                target_idx, target_origin = target_layout[target][1:]
                if target_idx is not None:
                    shape[target_idx] = target_origin

        if len(pop_targets) == 0:
            reshape_model(ov_model, shapes)
            break

    return ov_model


def load_ov_model(model_path: str, weight_path: Optional[str] = None, convert_dynamic: bool = False) -> Model:
    model_path = str(model_path)
    if model_path.startswith("omz://"):
        model_path = model_path.replace("omz://", "")
        assert model_path in AVAILABLE_OMZ_MODELS
        ov_ir_path = get_omz_model(model_path)
        model_path = ov_ir_path["model_path"]
        weight_path = ov_ir_path["weight_path"]

    if weight_path is None:
        weight_path = os.path.splitext(model_path)[0] + ".bin"

    if not os.path.exists(model_path):
        raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), model_path)
    if not os.path.exists(weight_path):
        raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), weight_path)

    ie = Core()
    ov_model = ie.read_model(model=model_path, weights=weight_path)

    if convert_dynamic:
        ov_model = to_dynamic_model(ov_model)

    return ov_model


def normalize_name(name: str) -> str:
    # ModuleDict does not allow '.' in module name string
    name = name.replace(".", "#")
    return name


def unnormalize_name(name: str) -> str:
    name = name.replace("#", ".")
    return name


def get_op_name(op: Node) -> str:
    op_name = op.get_friendly_name()
    op_name = normalize_name(op_name)
    return op_name


def convert_op_to_torch(op: Node):

    from .ops import OPS

    op_type = op.get_type_name()
    op_version = op.get_version()

    try:
        torch_module = OPS.get_by_type_version(op_type, op_version).from_ov(op)
    except Exception as e:
        logger.error(e)
        logger.error(op_type)
        logger.error(op_version)
        logger.error(op.get_attributes())
        raise e

    return torch_module


def convert_op_to_torch_module(target_op: Node):
    from .ops.modules import OperationModule

    dependent_modules = []
    for in_port in target_op.inputs():
        out_port = in_port.get_source_output()
        parent = out_port.get_node()

        parent_type = parent.get_type_name()
        if parent_type == "Constant":
            dependent_modules.append(convert_op_to_torch(parent))
        else:
            dependent_modules.append(None)
    module = convert_op_to_torch(target_op)
    module = OperationModule(module, dependent_modules)
    return module



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import os

import numpy as np
import pytest
import torch
from mmcls.models import build_classifier
from mmdet.models import build_detector

from otx.algorithms.classification.tasks import ClassificationInferenceTask  # noqa
from otx.cli.registry import Registry
from otx.mpa.det.stage import DetectionStage  # noqa
from otx.mpa.modules.hooks.recording_forward_hooks import (
    DetSaliencyMapHook,
    ReciproCAMHook,
)
from otx.mpa.utils.config_utils import MPAConfig
from tests.test_suite.e2e_test_system import e2e_pytest_unit

templates_cls = Registry("otx/algorithms").filter(task_type="CLASSIFICATION").templates
templates_cls_ids = [template.model_template_id for template in templates_cls]

templates_det = Registry("otx/algorithms").filter(task_type="DETECTION").templates
templates_det_ids = [template.model_template_id for template in templates_det]


class TestExplainMethods:
    ref_saliency_vals_cls = {
        "EfficientNet-B0": np.array([153, 102, 188, 189, 199, 159, 170], dtype=np.uint8),
        "MobileNet-V3-large-1x": np.array([140, 82, 87, 81, 79, 117, 254], dtype=np.uint8),
        "EfficientNet-V2-S": np.array([92, 197, 220, 227, 222, 196, 77], dtype=np.uint8),
    }

    ref_saliency_shapes = {
        "ATSS": (2, 4, 4),
        "SSD": (81, 13, 13),
        "YOLOX": (80, 13, 13),
    }

    ref_saliency_vals_det = {
        "ATSS": np.array([80, 213, 255, 124], dtype=np.uint8),
        "YOLOX": np.array([99, 38, 42, 57, 50, 71, 75, 81, 75, 61, 70, 5, 177], dtype=np.uint8),
        "SSD": np.array([191, 133, 160, 60, 63, 50, 51, 51, 57, 54, 83, 71, 148], dtype=np.uint8),
    }

    @e2e_pytest_unit
    @pytest.mark.parametrize("template", templates_cls, ids=templates_cls_ids)
    def test_saliency_map_cls(self, template):
        torch.manual_seed(0)

        base_dir = os.path.abspath(os.path.dirname(template.model_template_path))
        cfg_path = os.path.join(base_dir, "model.py")
        cfg = MPAConfig.fromfile(cfg_path)

        cfg.model.pop("task")
        model = build_classifier(cfg.model)
        model = model.eval()

        # img = torch.rand(2, 3, 224, 224) - 0.5
        img = torch.ones(2, 3, 224, 224) - 0.5
        data = {"img_metas": {}, "img": img}

        with ReciproCAMHook(model) as rcam_hook:
            with torch.no_grad():
                _ = model(return_loss=False, **data)
        saliency_maps = rcam_hook.records

        assert len(saliency_maps) == 2
        assert saliency_maps[0].ndim == 3
        assert saliency_maps[0].shape == (1000, 7, 7)
        assert (saliency_maps[0][0][0] == self.ref_saliency_vals_cls[template.name]).all()

    @e2e_pytest_unit
    @pytest.mark.parametrize("template", templates_det, ids=templates_det_ids)
    def test_saliency_map_det(self, template):
        torch.manual_seed(0)

        base_dir = os.path.abspath(os.path.dirname(template.model_template_path))
        cfg_path = os.path.join(base_dir, "model.py")
        cfg = MPAConfig.fromfile(cfg_path)

        model = build_detector(cfg.model)
        model = model.eval()

        # img = torch.rand(2, 3, 416, 416) - 0.5
        img = torch.ones(2, 3, 416, 416) - 0.5
        img_metas = [
            {
                "img_shape": (416, 416, 3),
                "scale_factor": np.array([1.1784703, 0.832, 1.1784703, 0.832], dtype=np.float32),
            },
            {
                "img_shape": (416, 416, 3),
                "scale_factor": np.array([1.1784703, 0.832, 1.1784703, 0.832], dtype=np.float32),
            },
        ]
        data = {"img_metas": [img_metas], "img": [img]}

        with DetSaliencyMapHook(model) as det_hook:
            with torch.no_grad():
                _ = model(return_loss=False, rescale=True, **data)
        saliency_maps = det_hook.records

        assert len(saliency_maps) == 2
        assert saliency_maps[0].ndim == 3
        assert saliency_maps[0].shape == self.ref_saliency_shapes[template.name]
        assert (saliency_maps[0][0][0] == self.ref_saliency_vals_det[template.name]).all()



"""Data Pipeline of Semi-SL model for Classification Task."""
# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

# pylint: disable=invalid-name
__img_norm_cfg = dict(mean=[123.675, 116.28, 103.53], std=[58.395, 57.12, 57.375], to_rgb=True)
__resize_target_size = 224

__common_pipeline = [
    dict(type="Resize", size=__resize_target_size),
    dict(type="RandomFlip", flip_prob=0.5, direction="horizontal"),
    dict(type="AugMixAugment", config_str="augmix-m5-w3"),
    dict(type="RandomRotate", p=0.35, angle=(-10, 10)),
]

__strong_pipeline = [
    dict(type="MPARandAugment", n=8, m=10),
]

__train_pipeline = [
    *__common_pipeline,
    dict(type="PILImageToNDArray", keys=["img"]),
    dict(type="Normalize", **__img_norm_cfg),
    dict(type="ImageToTensor", keys=["img"]),
    dict(type="ToTensor", keys=["gt_label"]),
    dict(type="Collect", keys=["img", "gt_label"]),
]

__unlabeled_pipeline = [
    *__common_pipeline,
    dict(type="PostAug", keys=dict(img_strong=__strong_pipeline)),
    dict(type="PILImageToNDArray", keys=["img", "img_strong"]),
    dict(type="Normalize", **__img_norm_cfg),
    dict(type="ImageToTensor", keys=["img", "img_strong"]),
    dict(type="Collect", keys=["img", "img_strong"]),
]

__test_pipeline = [
    dict(type="Resize", size=__resize_target_size),
    dict(type="Normalize", **__img_norm_cfg),
    dict(type="ImageToTensor", keys=["img"]),
    dict(type="Collect", keys=["img"]),
]

__dataset_type = "OTXClsDataset"
__samples_per_gpu = 32
__workers_per_gpu = 2

data = dict(
    samples_per_gpu=__samples_per_gpu,
    workers_per_gpu=__workers_per_gpu,
    train=dict(type=__dataset_type, pipeline=__train_pipeline),
    unlabeled=dict(
        type=__dataset_type,
        pipeline=__unlabeled_pipeline,
    ),
    val=dict(type=__dataset_type, test_mode=True, pipeline=__test_pipeline),
    test=dict(type=__dataset_type, test_mode=True, pipeline=__test_pipeline),
)



"""Sample Code of otx training for instance segmentation."""

# Copyright (C) 2021-2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import argparse
import sys

import cv2
import numpy as np
from mmcv.utils import get_logger

from otx.algorithms.common.utils import get_task_class
from otx.api.configuration.helper import create
from otx.api.entities.annotation import (
    Annotation,
    AnnotationSceneEntity,
    AnnotationSceneKind,
)
from otx.api.entities.dataset_item import DatasetItemEntity
from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.image import Image
from otx.api.entities.inference_parameters import InferenceParameters
from otx.api.entities.label import Domain, LabelEntity
from otx.api.entities.label_schema import LabelSchemaEntity
from otx.api.entities.model import ModelEntity
from otx.api.entities.model_template import parse_model_template
from otx.api.entities.optimization_parameters import OptimizationParameters
from otx.api.entities.resultset import ResultSetEntity
from otx.api.entities.scored_label import ScoredLabel
from otx.api.entities.shapes.polygon import Point, Polygon
from otx.api.entities.subset import Subset
from otx.api.entities.task_environment import TaskEnvironment
from otx.api.usecases.tasks.interfaces.export_interface import ExportType
from otx.api.usecases.tasks.interfaces.optimization_interface import OptimizationType

logger = get_logger(name="mmdet")

# pylint: disable=too-many-locals, too-many-statements


def parse_args():
    """Parse function for getting model template & check export."""
    parser = argparse.ArgumentParser(description="Sample showcasing the new API")
    parser.add_argument("template_file_path", help="path to template file")
    parser.add_argument("--export", action="store_true")
    return parser.parse_args()


colors = dict(red=(255, 0, 0), green=(0, 255, 0))


def load_test_dataset(data_type, task_type=Domain.INSTANCE_SEGMENTATION):
    """Load Sample dataset for Instance_segmentation."""

    def gen_circle_image(resolution):
        width, height = resolution
        image = np.full([height, width, 3], fill_value=255, dtype=np.uint8)
        gt_label = np.full([height, width, 1], fill_value=0, dtype=np.uint8)
        cv2.circle(image, (int(height / 2), int(width / 2)), 90, (0, 0, 255), -1)
        cv2.circle(gt_label, (int(height / 2), int(width / 2)), 90, 1, -1)
        return (image, gt_label)

    def gen_rect_image(resolution):
        width, height = resolution
        image = np.full([height, width, 3], fill_value=255, dtype=np.uint8)
        gt_label = np.full([height, width, 1], fill_value=0, dtype=np.uint8)
        cv2.rectangle(image, (int(height * 0.1), int(width * 0.1)), (int(height / 2), int(width / 2)), (0, 255, 0), -1)
        cv2.rectangle(gt_label, (int(height * 0.1), int(width * 0.1)), (int(height / 2), int(width / 2)), 2, -1)
        return (image, gt_label)

    labels = [
        LabelEntity(name="circle", domain=task_type, id=1),  # OLD class
        LabelEntity(name="rect", domain=task_type, id=2),
    ]

    def get_image(split_type, subset, label_id):
        ignored_labels = []
        height, width = 1280, 720
        if label_id == 1:
            image, gt_label = gen_circle_image((height, width))
            if split_type == "new" and subset == Subset.TRAINING:
                ignored_labels = [LabelEntity(name="rect", domain=Domain.INSTANCE_SEGMENTATION, id=2)]
        else:
            image, gt_label = gen_rect_image((height, width))

        height, width = gt_label.shape[:2]
        label_mask = gt_label == label_id
        label_index_map = label_mask.astype(np.uint8)
        contours, hierarchies = cv2.findContours(label_index_map, cv2.RETR_CCOMP, cv2.CHAIN_APPROX_SIMPLE)

        for contour, hierarchy in zip(contours, hierarchies[0]):
            if hierarchy[3] != -1:
                continue

            contour = list(contour)
            if len(contour) <= 2:
                continue

            points = [Point(x=point[0][0] / width, y=point[0][1] / height) for point in contour]

        return DatasetItemEntity(
            media=Image(data=image),
            annotation_scene=AnnotationSceneEntity(
                annotations=[Annotation(Polygon(points=points), labels=[ScoredLabel(label=labels[label_id - 1])])],
                kind=AnnotationSceneKind.ANNOTATION,
            ),
            subset=subset,
            ignored_labels=ignored_labels,
        )

    old_train = [
        get_image("old", Subset.TRAINING, 1),
        get_image("old", Subset.TRAINING, 1),
        get_image("old", Subset.TRAINING, 1),
        get_image("old", Subset.TRAINING, 1),
        get_image("old", Subset.TRAINING, 1),
        get_image("old", Subset.TRAINING, 1),
        get_image("old", Subset.TRAINING, 1),
        get_image("old", Subset.TRAINING, 1),
    ]

    old_val = [
        get_image("old", Subset.VALIDATION, 1),
        get_image("old", Subset.VALIDATION, 1),
        get_image("old", Subset.VALIDATION, 1),
        get_image("old", Subset.VALIDATION, 1),
        get_image("old", Subset.VALIDATION, 1),
        get_image("old", Subset.VALIDATION, 1),
        get_image("old", Subset.VALIDATION, 1),
        get_image("old", Subset.VALIDATION, 1),
    ]

    new_train = [
        get_image("new", Subset.TRAINING, 1),
        get_image("new", Subset.TRAINING, 1),
        get_image("new", Subset.TRAINING, 1),
        get_image("new", Subset.TRAINING, 1),
        get_image("new", Subset.TRAINING, 2),
        get_image("new", Subset.TRAINING, 2),
        get_image("new", Subset.TRAINING, 2),
        get_image("new", Subset.TRAINING, 2),
    ]

    new_val = [
        get_image("new", Subset.VALIDATION, 1),
        get_image("new", Subset.VALIDATION, 1),
        get_image("new", Subset.VALIDATION, 1),
        get_image("new", Subset.VALIDATION, 1),
        get_image("new", Subset.VALIDATION, 2),
        get_image("new", Subset.VALIDATION, 2),
        get_image("new", Subset.VALIDATION, 2),
        get_image("new", Subset.VALIDATION, 2),
    ]

    new_test = [
        get_image("old", Subset.TESTING, 1),
        get_image("old", Subset.TESTING, 1),
        get_image("old", Subset.TESTING, 1),
        get_image("old", Subset.TESTING, 1),
        get_image("old", Subset.TESTING, 1),
        get_image("old", Subset.TESTING, 1),
        get_image("old", Subset.TESTING, 1),
        get_image("old", Subset.TESTING, 1),
        get_image("new", Subset.TESTING, 1),
        get_image("new", Subset.TESTING, 1),
        get_image("new", Subset.TESTING, 2),
        get_image("new", Subset.TESTING, 2),
        get_image("new", Subset.TESTING, 2),
        get_image("new", Subset.TESTING, 2),
        get_image("new", Subset.TESTING, 2),
        get_image("new", Subset.TESTING, 2),
        get_image("new", Subset.TESTING, 2),
        get_image("new", Subset.TESTING, 2),
        get_image("new", Subset.TESTING, 2),
        get_image("new", Subset.TESTING, 2),
    ]

    old = old_train + old_val
    new = new_train + new_val
    if data_type == "old":
        return DatasetEntity(old * 5), [labels[0]]
    if data_type == "test":
        return DatasetEntity(new_test * 5), labels
    return DatasetEntity((old * 5 + new * 3)), labels


def main(args):
    """Main function of Detection Sample."""
    logger.info("[SL] Train initial model with OLD dataset")
    model_template = parse_model_template(args.template_file_path)
    task_type = model_template.task_type.domain

    logger.info("Train initial model with OLD dataset")
    dataset, labels_list = load_test_dataset("old", task_type)
    labels_schema = LabelSchemaEntity.from_labels(labels_list)

    logger.info(f"Train dataset: {len(dataset.get_subset(Subset.TRAINING))} items")
    logger.info(f"Validation dataset: {len(dataset.get_subset(Subset.VALIDATION))} items")

    logger.info("Set hyperparameters")
    params = create(model_template.hyper_parameters.data)
    params.learning_parameters.num_iters = 5
    params.learning_parameters.learning_rate = 0.01
    params.learning_parameters.learning_rate_warmup_iters = 1
    params.learning_parameters.batch_size = 4

    logger.info("Setup environment")
    environment = TaskEnvironment(
        model=None, hyper_parameters=params, label_schema=labels_schema, model_template=model_template
    )

    logger.info("Create base Task")
    task_impl_path = model_template.entrypoints.base
    task_cls = get_task_class(task_impl_path)
    task = task_cls(task_environment=environment)

    logger.info("Train model")
    initial_model = ModelEntity(
        dataset,
        environment.get_model_configuration(),
    )
    task.train(dataset, initial_model)

    logger.info("Class-incremental learning with OLD + NEW dataset")
    dataset, labels_list = load_test_dataset("new", task_type)
    labels_schema = LabelSchemaEntity.from_labels(labels_list)

    logger.info(f"Train dataset: {len(dataset.get_subset(Subset.TRAINING))} items")
    logger.info(f"Validation dataset: {len(dataset.get_subset(Subset.VALIDATION))} items")

    logger.info("Load model template")
    model_template = parse_model_template(args.template_file_path)

    logger.info("Set hyperparameters")
    params = create(model_template.hyper_parameters.data)
    params.learning_parameters.num_iters = 5
    params.learning_parameters.learning_rate = 0.015
    params.learning_parameters.learning_rate_warmup_iters = 1
    params.learning_parameters.batch_size = 4

    logger.info("Setup environment")
    environment = TaskEnvironment(
        model=initial_model, hyper_parameters=params, label_schema=labels_schema, model_template=model_template
    )

    logger.info("Create base Task")
    task_impl_path = model_template.entrypoints.base
    task_cls = get_task_class(task_impl_path)
    task = task_cls(task_environment=environment)

    logger.info("Train model")
    output_model = ModelEntity(
        dataset,
        environment.get_model_configuration(),
    )
    task.train(dataset, output_model)

    logger.info("Get predictions on the test set")
    testset, _ = load_test_dataset("test", task_type)
    validation_dataset = testset.get_subset(Subset.TESTING)
    predicted_validation_dataset = task.infer(
        validation_dataset.with_empty_annotations(), InferenceParameters(is_evaluation=True)
    )
    resultset = ResultSetEntity(
        model=output_model,
        ground_truth_dataset=validation_dataset,
        prediction_dataset=predicted_validation_dataset,
    )
    logger.info("Estimate quality on validation set")
    task.evaluate(resultset)
    logger.info(str(resultset.performance))

    if args.export:
        logger.info("Export model")
        exported_model = ModelEntity(
            dataset,
            environment.get_model_configuration(),
        )
        task.export(ExportType.OPENVINO, exported_model)

        logger.info("Create OpenVINO Task")
        environment.model = exported_model
        openvino_task_impl_path = model_template.entrypoints.openvino
        openvino_task_cls = get_task_class(openvino_task_impl_path)
        openvino_task = openvino_task_cls(environment)

        logger.info("Get predictions on the validation set")
        predicted_validation_dataset = openvino_task.infer(
            validation_dataset.with_empty_annotations(), InferenceParameters(is_evaluation=True)
        )

        resultset = ResultSetEntity(
            model=output_model,
            ground_truth_dataset=validation_dataset,
            prediction_dataset=predicted_validation_dataset,
        )
        logger.info("Estimate quality on validation set")
        openvino_task.evaluate(resultset)
        logger.info(str(resultset.performance))

        logger.info("Run POT optimization")
        optimized_model = ModelEntity(
            dataset,
            environment.get_model_configuration(),
        )
        openvino_task.optimize(OptimizationType.POT, dataset, optimized_model, OptimizationParameters())

        logger.info("Get predictions on the validation set")
        predicted_validation_dataset = openvino_task.infer(
            validation_dataset.with_empty_annotations(), InferenceParameters(is_evaluation=True)
        )
        resultset = ResultSetEntity(
            model=optimized_model,
            ground_truth_dataset=validation_dataset,
            prediction_dataset=predicted_validation_dataset,
        )
        logger.info("Performance of optimized model:")
        openvino_task.evaluate(resultset)
        logger.info(str(resultset.performance))

        logger.info("Running the NNCF optimization")
        environment.model = output_model
        nncf_task_impl_path = model_template.entrypoints.nncf
        nncf_task_cls = get_task_class(nncf_task_impl_path)
        nncf_task = nncf_task_cls(environment)

        optimized_model = ModelEntity(
            dataset,
            configuration=environment.get_model_configuration(),
        )
        nncf_task.optimize(OptimizationType.NNCF, dataset, optimized_model)

        logger.info("Inferring the optimised model on the validation set.")
        predicted_validation_dataset = nncf_task.infer(
            validation_dataset.with_empty_annotations(),
            InferenceParameters(is_evaluation=True),
        )
        resultset = ResultSetEntity(
            model=optimized_model,
            ground_truth_dataset=validation_dataset,
            prediction_dataset=predicted_validation_dataset,
        )

        logger.info("Evaluating the optimized model on the validation set.")
        nncf_task.evaluate(resultset)
        logger.info(str(resultset.performance))

        logger.info("Exporting the model.")
        exported_model = ModelEntity(
            train_dataset=dataset,
            configuration=environment.get_model_configuration(),
        )
        nncf_task.export(ExportType.OPENVINO, exported_model)
        environment.model = exported_model

        logger.info("Creating the OpenVINO Task.")
        environment.model = exported_model
        openvino_task_impl_path = model_template.entrypoints.openvino
        nncf_openvino_task_cls = get_task_class(openvino_task_impl_path)
        nncf_openvino_task = nncf_openvino_task_cls(environment)

        logger.info("Inferring the exported model on the validation set.")
        predicted_validation_dataset = nncf_openvino_task.infer(
            validation_dataset.with_empty_annotations(),
            InferenceParameters(is_evaluation=True),
        )

        logger.info("Evaluating the exported model on the validation set.")
        resultset = ResultSetEntity(
            model=exported_model,
            ground_truth_dataset=validation_dataset,
            prediction_dataset=predicted_validation_dataset,
        )
        nncf_openvino_task.evaluate(resultset)
        logger.info(str(resultset.performance))


if __name__ == "__main__":
    sys.exit(main(parse_args()) or 0)



# Copyright (C) 2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from abc import abstractmethod
from collections import Counter, OrderedDict
from copy import deepcopy
from typing import Callable, Dict, List, Optional, Type

from .e2e_test_system import DataCollector
from .logging import get_logger
from .training_tests_actions import BaseOTXTestAction
from .training_tests_stage import OTXTestStage, OTXTestStagesStorageInterface, Validator

logger = get_logger()


def _get_duplications(arr):
    c = Counter(arr)
    dups = [k for k, v in c.items() if v > 1]
    return dups


def _str_dict_with_shortened_vals(d, max_len=200):
    assert isinstance(d, dict)
    if not d:
        return "{}"

    def _shorten(v):
        sv = str(v)
        if len(sv) <= max_len:
            return sv
        return sv[:max_len] + "..."

    s = "\n".join(f"{k}: {_shorten(v)}," for k, v in d.items())
    s = "\n    ".join(s.split("\n"))
    s = "{\n    " + s + "\n}"
    return s


class OTXTestCaseInterface(OTXTestStagesStorageInterface):
    @classmethod
    @abstractmethod
    def get_list_of_test_stages(cls):
        raise NotImplementedError("The method get_list_of_test_stages is not implemented")

    @abstractmethod
    def run_stage(
        self,
        stage_name: str,
        data_collector: DataCollector,
        cur_test_expected_metrics_callback: Optional[Callable[[], Dict]],
    ):
        raise NotImplementedError("The method run_stage is not implemented")


def generate_otx_integration_test_case_class(
    test_actions_classes: List[Type[BaseOTXTestAction]],
) -> Type:
    test_actions_classes = deepcopy(test_actions_classes)

    # check names' duplication
    classes_names = [action_cls._name for action_cls in test_actions_classes]
    name_dups = _get_duplications(classes_names)
    if name_dups:
        raise ValueError(f"Wrong input: there are duplications in names of actions; duplications = {name_dups}")

    class _OTXIntegrationTestCase(OTXTestCaseInterface):
        _TEST_STAGES = [action_cls._name for action_cls in test_actions_classes]

        @classmethod
        def get_list_of_test_stages(cls):
            return deepcopy(cls._TEST_STAGES)

        def __init__(self, params_factories_for_test_actions: Dict[str, Callable[[], Dict]]):
            logger.debug("initialization of test case: begin")
            self._stages = OrderedDict()
            for action_cls in test_actions_classes:
                logger.debug(f"initialization of test case: action_cls={action_cls}")

                cur_name = action_cls._name
                assert cur_name is not None
                cur_params_factory = params_factories_for_test_actions.get(cur_name)
                if cur_params_factory is not None:
                    logger.debug("initialization of test case: calling params factory")
                    cur_params = cur_params_factory()
                else:
                    cur_params = {}

                assert isinstance(cur_params, dict), f"Wrong params received from factory: {cur_params}"
                short_params_str = _str_dict_with_shortened_vals(cur_params)
                logger.info(f"initialization of test case: add action '{cur_name}' " f"with params={short_params_str}")

                cur_action = action_cls(**cur_params)

                # Note that `self` is used as stages_storage for OTXTestStage below
                cur_stage = OTXTestStage(action=cur_action, stages_storage=self)
                self._stages[cur_name] = cur_stage

            assert list(self._stages.keys()) == list(self._TEST_STAGES)

            # test results should be kept between stages
            self.test_results_storage: OrderedDict = OrderedDict()
            logger.debug("initialization of test case: end")

        # implementation of method from OTXTestStagesStorageInterface
        def get_stage(self, name: str) -> "OTXTestStage":
            return self._stages[name]

        def run_stage(
            self,
            stage_name: str,
            data_collector: DataCollector,
            cur_test_expected_metrics_callback: Optional[Callable[[], Dict]],
        ):
            assert stage_name in self._TEST_STAGES, f"Wrong stage_name {stage_name}"
            validator = Validator(cur_test_expected_metrics_callback)
            self._stages[stage_name].run_once(data_collector, self.test_results_storage, validator)

    return _OTXIntegrationTestCase



"""Anomaly Dataset Utils."""

# Copyright (C) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

from typing import Dict, List, Optional, Union

import numpy as np
import torch
from anomalib.data.base.datamodule import collate_fn
from anomalib.data.utils.transform import get_transforms
from omegaconf import DictConfig, ListConfig
from pytorch_lightning.core.datamodule import LightningDataModule
from torch import Tensor
from torch.utils.data import DataLoader, Dataset

from otx.algorithms.anomaly.adapters.anomalib.logger import get_logger
from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.model_template import TaskType
from otx.api.entities.shapes.polygon import Polygon
from otx.api.entities.shapes.rectangle import Rectangle
from otx.api.entities.subset import Subset
from otx.api.utils.dataset_utils import (
    contains_anomalous_images,
    split_local_global_dataset,
)
from otx.api.utils.segmentation_utils import mask_from_dataset_item

logger = get_logger(__name__)


class OTXAnomalyDataset(Dataset):
    """Anomaly Dataset Adaptor.

    This class converts OTX Dataset into Anomalib dataset that
    is a sub-class of Vision Dataset.

    Args:
        config (Union[DictConfig, ListConfig]): Anomalib config
        dataset (DatasetEntity): [description]: OTX SDK Dataset

    Example:
        >>> from tests.helpers.dataset import OTXAnomalyDatasetGenerator
        >>> from otx.utils.data import AnomalyDataset

        >>> dataset_generator = OTXAnomalyDatasetGenerator()
        >>> dataset = dataset_generator.generate()
        >>> anomaly_dataset = AnomalyDataset(config=config, dataset=dataset)
        >>> anomaly_dataset[0]["image"].shape
        torch.Size([3, 256, 256])
    """

    def __init__(self, config: Union[DictConfig, ListConfig], dataset: DatasetEntity, task_type: TaskType):
        self.config = config
        self.dataset = dataset
        self.task_type = task_type

        # TODO: distinguish between train and val config here
        self.transform = get_transforms(
            config=config.dataset.transform_config.train, image_size=tuple(config.dataset.image_size), to_tensor=True
        )

    def __len__(self) -> int:
        """Get size of the dataset.

        Returns:
            int: Size of the dataset.
        """
        return len(self.dataset)

    def __getitem__(self, index: int) -> Dict[str, Union[int, Tensor]]:
        """Get dataset item.

        Args:
            index (int): Index of the dataset sample.

        Raises:
            ValueError: When the task type is not supported.

        Returns:
            Dict[str, Union[int, Tensor]]: Dataset item.
        """
        dataset_item = self.dataset[index]
        item: Dict[str, Union[int, Tensor]] = {}
        item = {"index": index}
        if self.task_type == TaskType.ANOMALY_CLASSIFICATION:
            # Detection currently relies on image labels only, meaning it'll use image
            #   threshold to find the predicted bounding boxes.
            item["image"] = self.transform(image=dataset_item.numpy)["image"]
        elif self.task_type == TaskType.ANOMALY_DETECTION:
            item["image"] = self.transform(image=dataset_item.numpy)["image"]
            item["boxes"] = torch.empty((0, 4))
            height, width = self.config.dataset.image_size
            boxes = []
            for annotation in dataset_item.get_annotations():
                if isinstance(annotation.shape, Rectangle) and not Rectangle.is_full_box(annotation.shape):
                    boxes.append(
                        Tensor(
                            [
                                annotation.shape.x1 * width,
                                annotation.shape.y1 * height,
                                annotation.shape.x2 * width,
                                annotation.shape.y2 * height,
                            ]
                        )
                    )
                if boxes:
                    item["boxes"] = torch.stack(boxes)
        elif self.task_type == TaskType.ANOMALY_SEGMENTATION:
            if any((isinstance(annotation.shape, Polygon) for annotation in dataset_item.get_annotations())):
                mask = mask_from_dataset_item(dataset_item, dataset_item.get_shapes_labels()).squeeze()
            else:
                mask = np.zeros(dataset_item.numpy.shape[:2]).astype(np.int)
            pre_processed = self.transform(image=dataset_item.numpy, mask=mask)
            item["image"] = pre_processed["image"]
            item["mask"] = pre_processed["mask"]
        else:
            raise ValueError(f"Unsupported task type: {self.task_type}")

        if len(dataset_item.get_shapes_labels()) > 0 and dataset_item.get_shapes_labels()[0].is_anomalous:
            item["label"] = 1
        else:
            item["label"] = 0
        return item


class OTXAnomalyDataModule(LightningDataModule):
    """Anomaly DataModule.

    This class converts OTX Dataset into Anomalib dataset and stores
    train/val/test dataloaders.

    Args:
        config (Union[DictConfig, ListConfig]): Anomalib config
        dataset (DatasetEntity): OTX SDK Dataset

    Example:
        >>> from tests.helpers.dataset import OTXAnomalyDatasetGenerator
        >>> from otx.utils.data import AnomalyDataModule

        >>> dataset_generator = OTXAnomalyDatasetGenerator()
        >>> dataset = dataset_generator.generate()
        >>> data_module = OTXAnomalyDataModule(config=config, dataset=dataset)
        >>> i, data = next(enumerate(data_module.train_dataloader()))
        >>> data["image"].shape
        torch.Size([32, 3, 256, 256])
    """

    def __init__(self, config: Union[DictConfig, ListConfig], dataset: DatasetEntity, task_type: TaskType) -> None:
        super().__init__()
        self.config = config
        self.dataset = dataset
        self.task_type = task_type

        self.train_otx_dataset: DatasetEntity
        self.val_otx_dataset: DatasetEntity
        self.test_otx_dataset: DatasetEntity
        self.predict_otx_dataset: DatasetEntity

    def setup(self, stage: Optional[str] = None) -> None:
        """Setup Anomaly Data Module.

        Args:
            stage (Optional[str], optional): train/val/test stages.
                Defaults to None.
        """
        if not stage == "predict":
            self.summary()

        if stage == "fit" or stage is None:
            self.train_otx_dataset = self.dataset.get_subset(Subset.TRAINING)
            self.val_otx_dataset = self.dataset.get_subset(Subset.VALIDATION)

        if stage == "validate":
            self.val_otx_dataset = self.dataset.get_subset(Subset.VALIDATION)

        if stage == "test" or stage is None:
            self.test_otx_dataset = self.dataset.get_subset(Subset.TESTING)

        if stage == "predict":
            self.predict_otx_dataset = self.dataset

    def summary(self):
        """Print size of the dataset, number of anomalous images and number of normal images."""
        for subset in [Subset.TRAINING, Subset.VALIDATION, Subset.TESTING]:
            dataset = self.dataset.get_subset(subset)
            num_items = len(dataset)
            num_normal = len([item for item in dataset if not item.get_shapes_labels()[0].is_anomalous])
            num_anomalous = len([item for item in dataset if item.get_shapes_labels()[0].is_anomalous])
            logger.info(
                "'%s' subset size: Total '%d' images. Normal: '%d', images. Anomalous: '%d' images",
                subset,
                num_items,
                num_normal,
                num_anomalous,
            )

    def train_dataloader(
        self,
    ) -> Union[DataLoader, List[DataLoader], Dict[str, DataLoader]]:
        """Train Dataloader.

        Returns:
            Union[DataLoader, List[DataLoader], Dict[str, DataLoader]]: Train dataloader.
        """
        dataset = OTXAnomalyDataset(self.config, self.train_otx_dataset, self.task_type)
        return DataLoader(
            dataset,
            shuffle=False,
            batch_size=self.config.dataset.train_batch_size,
            num_workers=self.config.dataset.num_workers,
            collate_fn=collate_fn,
        )

    def val_dataloader(self) -> Union[DataLoader, List[DataLoader]]:
        """Validation Dataloader.

        Returns:
            Union[DataLoader, List[DataLoader]]: Validation Dataloader.
        """
        global_dataset, local_dataset = split_local_global_dataset(self.val_otx_dataset)
        logger.info(f"Global annotations: {len(global_dataset)}")
        logger.info(f"Local annotations: {len(local_dataset)}")
        if contains_anomalous_images(local_dataset):
            logger.info("Dataset contains polygon annotations. Passing masks to anomalib.")
            dataset = OTXAnomalyDataset(self.config, local_dataset, self.task_type)
        else:
            logger.info("Dataset does not contain polygon annotations. Not passing masks to anomalib.")
            dataset = OTXAnomalyDataset(self.config, global_dataset, TaskType.ANOMALY_CLASSIFICATION)
        return DataLoader(
            dataset,
            shuffle=False,
            batch_size=self.config.dataset.eval_batch_size,
            num_workers=self.config.dataset.num_workers,
            collate_fn=collate_fn,
        )

    def test_dataloader(self) -> Union[DataLoader, List[DataLoader]]:
        """Test Dataloader.

        Returns:
            Union[DataLoader, List[DataLoader]]: Test Dataloader.
        """
        dataset = OTXAnomalyDataset(self.config, self.test_otx_dataset, self.task_type)
        return DataLoader(
            dataset,
            shuffle=False,
            batch_size=self.config.dataset.test_batch_size,
            num_workers=self.config.dataset.num_workers,
            collate_fn=collate_fn,
        )

    def predict_dataloader(self) -> Union[DataLoader, List[DataLoader]]:
        """Predict Dataloader.

        Returns:
            Union[DataLoader, List[DataLoader]]: Predict Dataloader.
        """
        dataset = OTXAnomalyDataset(self.config, self.predict_otx_dataset, self.task_type)
        return DataLoader(
            dataset,
            shuffle=False,
            batch_size=self.config.dataset.eval_batch_size,
            num_workers=self.config.dataset.num_workers,
            collate_fn=collate_fn,
        )



import os

import pytest

from otx.mpa.seg.trainer import SegTrainer
from otx.mpa.utils.config_utils import MPAConfig
from tests.test_suite.e2e_test_system import e2e_pytest_unit
from tests.unit.algorithms.segmentation.test_helpers import (
    DEFAULT_RECIPE_CONFIG_PATH,
    DEFAULT_SEG_TEMPLATE_DIR,
)


class TestOTXSegTrainer:
    @pytest.fixture(autouse=True)
    def setup(self) -> None:
        cfg = MPAConfig.fromfile(DEFAULT_RECIPE_CONFIG_PATH)
        self.trainer = SegTrainer(name="", mode="train", config=cfg, common_cfg=None, index=0)
        self.model_cfg = MPAConfig.fromfile(os.path.join(DEFAULT_SEG_TEMPLATE_DIR, "model.py"))
        self.data_cfg = MPAConfig.fromfile(os.path.join(DEFAULT_SEG_TEMPLATE_DIR, "data_pipeline.py"))

    @e2e_pytest_unit
    def test_run(self, mocker):
        mocker.patch.object(SegTrainer, "configure_samples_per_gpu")
        mocker.patch.object(SegTrainer, "configure_fp16_optimizer")
        mocker.patch.object(SegTrainer, "configure_compat_cfg")
        mock_train_segmentor = mocker.patch("otx.mpa.seg.trainer.train_segmentor")

        self.trainer.run(self.model_cfg, "", self.data_cfg)
        mock_train_segmentor.assert_called_once()

    @e2e_pytest_unit
    def test_run_with_distributed(self, mocker):
        self.trainer._distributed = True
        mocker.patch.object(SegTrainer, "configure_samples_per_gpu")
        mocker.patch.object(SegTrainer, "configure_fp16_optimizer")
        mocker.patch.object(SegTrainer, "configure_compat_cfg")
        spy_cfg_dist = mocker.spy(SegTrainer, "_modify_cfg_for_distributed")
        mock_train_segmentor = mocker.patch("otx.mpa.seg.trainer.train_segmentor")

        self.trainer.run(self.model_cfg, "", self.data_cfg)
        spy_cfg_dist.assert_called_once()
        mock_train_segmentor.assert_called_once()



"""NNCF utils for mmseg."""
# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

# flake8: noqa

from .builder import build_nncf_segmentor
from .hooks import CustomstepLrUpdaterHook

__all__ = [
    "build_nncf_segmentor",
    "CustomstepLrUpdaterHook",
]



"""OTX Algorithms - Utils."""

# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from .labels_utils import get_multihead_class_info

__all__ = ["get_multihead_class_info"]



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import onnx


def remove_nodes_by_op_type(onnx_model, op_type):
    # TODO: support more nodes

    supported_op_types = ["Mark", "Conv", "Gemm"]
    assert op_type in supported_op_types

    target_nodes = []
    for node in onnx_model.graph.node:
        if node.op_type == op_type:
            target_nodes.append(node)

    output_names = [i.name for i in onnx_model.graph.output]

    for target_node in target_nodes:
        in_port = target_node.input[0]
        out_port = target_node.output[0]

        out_nodes = [node for node in onnx_model.graph.node if out_port in node.input]
        for out_node in out_nodes:
            for i, j in enumerate(out_node.input):
                if j == out_port:
                    out_node.input[i] = in_port

        if out_port in output_names:
            in_nodes = [node for node in onnx_model.graph.node if in_port in node.output]
            for in_node in in_nodes:
                for i, j in enumerate(in_node.output):
                    if j == in_port:
                        in_node.output[i] = out_port

        onnx_model.graph.node.remove(target_node)

    onnx.checker.check_model(onnx_model)
    return onnx_model


def prepare_onnx_for_openvino(in_path, out_path):
    onnx_model = onnx.load(in_path)
    onnx_model = remove_nodes_by_op_type(onnx_model, "Mark")
    onnx.checker.check_model(onnx_model)
    onnx.save(onnx_model, out_path)



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import torch
import torch.nn as nn
import torch.nn.functional as F


class OnnxLpNormalization(torch.autograd.Function):
    @staticmethod
    def forward(ctx, x, axis=0, p=2, eps=1e-12):
        denom = x.norm(2, axis, True).clamp_min(eps).expand_as(x)
        return x / denom

    @staticmethod
    def symbolic(g, x, axis=0, p=2, eps=1e-12):
        return g.op("LpNormalization", x, axis_i=int(axis), p_i=int(p))


def normalize(x, dim, p=2, eps=1e-12):
    if torch.onnx.is_in_onnx_export():
        return OnnxLpNormalization.apply(x, dim, p, eps)
    else:
        return F.normalize(x, dim=dim, p=p, eps=eps)


class Normalize(nn.Module):
    def __init__(self, dim=1, p=2, eps=1e-12):
        super().__init__()

        self.dim = dim
        self.p = p
        self.eps = eps

    def forward(self, x):
        return normalize(x, self.dim, self.p, self.eps)



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from mmseg.models.builder import MODELS

SCALAR_SCHEDULERS = MODELS


def build_scalar_scheduler(cfg, default_value=None):
    if cfg is None:
        if default_value is not None:
            assert isinstance(default_value, (int, float))
            cfg = dict(type="ConstantScalarScheduler", scale=float(default_value))
        else:
            return None
    elif isinstance(cfg, (int, float)):
        cfg = dict(type="ConstantScalarScheduler", scale=float(cfg))

    return SCALAR_SCHEDULERS.build(cfg)



"""
OTX components markers.
"""

# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#


try:
    from e2e.markers.mark_meta import MarkMeta
except ImportError:

    class MarkMeta:  # type: ignore[no-redef]
        """
        Empty marker.
        """


class OtxSdkComponent(MarkMeta):
    """
    OTX Component marker.
    """

    OTX_API = "otx.api"



"""This module contains the AutoHPOState Enum."""


# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from enum import Enum


class AutoHPOState(Enum):
    """Holds metadata related to automatic hyper parameter optimization (auto-HPO) for a single configurable parameter.

    It contains the following values:
        NOT_POSSIBLE  - This implies that the parameter cannot be optimized via auto-HPO
        POSSIBLE      - This implies that the parameter can potentially be optimized via
                        auto-HPO, but auto-HPO has not been carried out for this parameter
                        yet
        OPTIMIZED     - This implies that the parameter has been optimized via auto-HPO,
                        such that the current value of the parameter reflects it's optimal
                        value
        OVERRIDDEN    - This implies that the parameter has previously been optimized via
                        auto-HPO, but it's value has been manually overridden
    """

    NOT_POSSIBLE = "not_possible"
    POSSIBLE = "possible"
    OPTIMIZED = "optimized"
    OVERRIDDEN = "overridden"

    def __str__(self):
        """Retrieves the string representation of an instance of the Enum."""
        return self.value



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import math
from functools import partial

from torch.nn import functional as F


def get_dynamic_shape(op):
    shape = [str(i) for i in op.get_partial_shape()]
    for i, shape_ in enumerate(shape):
        try:
            shape_ = int(shape_)
        except ValueError:
            shape_ = -1
        shape[i] = shape_
    return shape


def get_torch_padding(pads_begin, pads_end, auto_pad, input_size, weight_size, stride, dilation=None):
    from .movements import PadV1

    if dilation is None:
        dilation = [1 for _ in input_size]

    if auto_pad == "valid":
        return 0
    elif auto_pad == "same_upper" or auto_pad == "same_lower":
        assert len(set(dilation)) == 1 and dilation[0] == 1
        pads_begin = []
        pads_end = []
        for input_size_, weight_size_, stride_, dilation_ in zip(input_size, weight_size, stride, dilation):
            out_size = math.ceil(input_size_ / stride_)
            padding_needed = max(0, (out_size - 1) * stride_ + weight_size_ - input_size_)
            padding_lhs = int(padding_needed / 2)
            padding_rhs = padding_needed - padding_lhs

            pads_begin.append(padding_lhs if auto_pad == "same_upper" else padding_rhs)
            pads_end.append(padding_rhs if auto_pad == "same_upper" else padding_lhs)
        pad = PadV1.get_torch_pad_dim(pads_begin, pads_end)
        return partial(F.pad, pad=pad, mode="constant", value=0)
    elif auto_pad == "explicit":
        pad = PadV1.get_torch_pad_dim(pads_begin, pads_end)
        return partial(F.pad, pad=pad, mode="constant", value=0)
    else:
        raise NotImplementedError



# Copyright (C) 2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import itertools
from abc import ABC, abstractmethod
from collections import OrderedDict
from copy import deepcopy
from pprint import pformat
from typing import Any, Dict, List, Optional, Type

from .logging import get_logger
from .training_test_case import (
    OTXTestCaseInterface,
    generate_otx_integration_test_case_class,
)
from .training_tests_actions import get_default_test_action_classes
from .training_tests_common import DEFAULT_FIELD_VALUE_FOR_USING_IN_TEST

logger = get_logger()


class OTXTrainingTestInterface(ABC):
    """
    The interface for all OTX training tests
    (both reallife training tests and integration training tests)
    """

    @classmethod
    @abstractmethod
    def get_list_of_tests(cls, usecase: Optional[str] = None):
        raise NotImplementedError("The method is not implemented")


class OTXTestCreationParametersInterface(ABC):
    """
    The interface for classes that gives parameters for creating
    OTX training tests.
    It is used as the input value for OTXTestHelper class that
    makes most part of functionality for the training tests.
    """

    @abstractmethod
    def test_bunches(self) -> List[Dict[str, Any]]:
        """
        The method should return test bunches struct.
        It should be a list of dicts, each dict contains info on one "test bunch",
        e.g.
            ```
            [
                dict(
                    model_name=[
                       'gen3_mobilenetV2_SSD',
                       'gen3_mobilenetV2_ATSS',
                       'gen3_resnet50_VFNet',
                    ],
                    dataset_name='dataset1_tiled_shortened_500_A',
                    usecase='precommit',
                ),
                ...
            ]
            ```
        Note that the dict-s are passed to the tests as is through the parameter 'test_parameters'
        -- see the method OTXTestHelper.get_list_of_tests below and the fixture
        fixtures.current_test_parameters_fx.
        """
        raise NotImplementedError("The method is not implemented")

    @abstractmethod
    def test_case_class(self) -> Type[OTXTestCaseInterface]:
        """
        The method returns a class that will be used as a Test Case class
        for training tests.
        Note that it should return a class itself (not an instance of the class).

        Typically OTX Test Case class should be generated by the function
        training_test_case.generate_otx_integration_test_case_class.

        Note that the function receives as the parameter the list of action
        classes -- see the function
        training_tests_actions.get_default_test_action_classes,
        it returns the default test action classes.
        """
        raise NotImplementedError("The method is not implemented")

    @abstractmethod
    def short_test_parameters_names_for_generating_id(self) -> OrderedDict:
        """
        The method returns an OrderedDict that is used for generating string id-s of tests
        by test parameters, received from test bunches dicts:
        * keys of the OrderedDict should be the string keys of test bunches dict-s that should be
          used for generating id-s
        * values of the OrderedDict should be the strings that will be used as names of test parameters

        See the function OTXTestHelper._generate_test_id below.
        """
        raise NotImplementedError("The method is not implemented")

    @abstractmethod
    def test_parameters_defining_test_case_behavior(self) -> List[str]:
        """
        The method returns a list of strings -- names of the test parameters
        (i.e. keys of test bunches dicts) that define test case behavior.

        When several test cases are handled, if the next test has these parameters
        the same as for the previous test, the test case class is re-used for the next test.
        This allows re-using the result of previous test stages in the next test stages.
        """
        raise NotImplementedError("The method is not implemented")

    @abstractmethod
    def default_test_parameters(self) -> Dict[str, Any]:
        """
        The method returns a dict that points for test parameters
        the default values.

        If some dict in test bunches does not have a field that is pointed
        in the dict returned by default_test_parameters, the value for the field is
        set by the default value.
        """
        raise NotImplementedError("The method is not implemented")


class DefaultOTXTestCreationParametersInterface(OTXTestCreationParametersInterface):
    """
    The default implementation of some of the parameters for creation training tests.
    """

    def test_case_class(self) -> Type[OTXTestCaseInterface]:
        return generate_otx_integration_test_case_class(get_default_test_action_classes())

    def short_test_parameters_names_for_generating_id(self) -> OrderedDict:
        DEFAULT_SHORT_TEST_PARAMETERS_NAMES_FOR_GENERATING_ID = OrderedDict(
            [
                ("test_stage", "ACTION"),
                ("model_name", "model"),
                ("dataset_name", "dataset"),
                ("num_training_iters", "num_iters"),
                ("batch_size", "batch"),
                ("usecase", "usecase"),
            ]
        )
        return deepcopy(DEFAULT_SHORT_TEST_PARAMETERS_NAMES_FOR_GENERATING_ID)

    def test_parameters_defining_test_case_behavior(self) -> List[str]:
        DEFAULT_TEST_PARAMETERS_DEFINING_IMPL_BEHAVIOR = [
            "model_name",
            "dataset_name",
            "num_training_iters",
            "batch_size",
        ]
        return deepcopy(DEFAULT_TEST_PARAMETERS_DEFINING_IMPL_BEHAVIOR)

    def default_test_parameters(self) -> Dict[str, Any]:
        DEFAULT_TEST_PARAMETERS = {
            "num_training_iters": 1,
            "batch_size": 2,
        }
        return deepcopy(DEFAULT_TEST_PARAMETERS)


class OTXTestHelper:
    """
    The main class helping creating OTX training tests.

    The instance of this class (with proper test_creation_parameters) should be
    created as a static field for the test class making training tests
    -- the class should be derived from the interface OTXTrainingTestInterface
    and forward the call of the class method get_list_of_tests to the helper
    (see method get_list_of_tests below).

    The most important method of the class are
    * get_list_of_tests -- allows pytest trick generating test parameters for
                           the test class
    * get_test_case -- gets an instance of the test case class for the current test parameters,
                       allows re-using the instance between several tests.
    """

    class _Cache:
        def __init__(self):
            self._cache_parameters = {}
            self._cached_value = None

        def get(self):
            return self._cached_value

        def set(self, params, value):
            logger.debug(f"cache.set new value for parameters {params}")
            self._cache_parameters = deepcopy(params)
            self._cached_value = value

        def has_same_params(self, params):
            res = self._cache_parameters == params
            res_str = "==" if res else "!="
            logger.debug(
                f"cache.has_same_params: " f"cache_parameters={self._cache_parameters} {res_str} {params}, res={res}"
            )
            return res

    def __init__(self, test_creation_parameters: OTXTestCreationParametersInterface):
        assert isinstance(test_creation_parameters, OTXTestCreationParametersInterface)

        self.test_case_class = test_creation_parameters.test_case_class()
        self.test_bunches = test_creation_parameters.test_bunches()

        self.short_test_parameters_names_for_generating_id = (
            test_creation_parameters.short_test_parameters_names_for_generating_id()
        )
        self.test_parameters_defining_test_case_behavior = (
            test_creation_parameters.test_parameters_defining_test_case_behavior()
        )
        self.default_test_parameters = test_creation_parameters.default_test_parameters()

        self._cache = OTXTestHelper._Cache()

        assert issubclass(self.test_case_class, OTXTestCaseInterface)
        assert isinstance(self.short_test_parameters_names_for_generating_id, OrderedDict)
        assert all(
            isinstance(k, str) and isinstance(v, str)
            for k, v in self.short_test_parameters_names_for_generating_id.items()
        )
        assert "test_stage" in self.short_test_parameters_names_for_generating_id
        assert "model_name" in self.short_test_parameters_names_for_generating_id
        assert "dataset_name" in self.short_test_parameters_names_for_generating_id
        assert "usecase" in self.short_test_parameters_names_for_generating_id

        assert isinstance(self.test_parameters_defining_test_case_behavior, list)
        assert all(isinstance(s, str) for s in self.test_parameters_defining_test_case_behavior)

        assert isinstance(self.default_test_parameters, dict)
        assert all(isinstance(k, str) for k in self.default_test_parameters.keys())

    def _get_list_of_test_stages(self):
        return self.test_case_class.get_list_of_test_stages()

    def _fill_test_parameters_default_values(self, test_parameters):
        for key, default_val in self.default_test_parameters.items():
            val = test_parameters.get(key)
            if val is None or val == DEFAULT_FIELD_VALUE_FOR_USING_IN_TEST:
                test_parameters[key] = default_val

    def _generate_test_id(self, test_parameters):
        param_name_to_short_name = self.short_test_parameters_names_for_generating_id
        id_parts = []
        for par_name, short_par_name in param_name_to_short_name.items():
            if par_name not in test_parameters:
                raise ValueError(
                    f"Test parameters do not contain key '{par_name}', " f"test_parameters={pformat(test_parameters)}"
                )
            id_parts.append(f"{short_par_name}-{test_parameters[par_name]}")

        return ",".join(id_parts)

    def get_list_of_tests(self, usecase: Optional[str] = None):
        """
        The functions generates the lists of values for the tests from the field test_bunches of the class.

        The function returns two lists
        * argnames -- a tuple with names of the test parameters, at the moment it is
                      a one-element tuple with the parameter name "test_parameters"
        * argvalues -- list of tuples, each tuple has the same len as argname tuple,
                       at the moment it is a one-element tuple with the dict `test_parameters`
                       that stores the parameters of the test
        * ids -- list of strings with ids corresponding the parameters of the tests
                 each id is a string generated from the corresponding test_parameters
                 value -- see the functions _generate_test_id

        The lists argvalues and ids will have the same length.

        If the parameter `usecase` is set, it makes filtering by usecase field of test bunches.
        """
        test_bunches = self.test_bunches
        assert all(isinstance(el, dict) for el in test_bunches)

        argnames = ("test_parameters",)
        argvalues = []
        ids = []
        for el in test_bunches:
            el_model_name = el.get("model_name")
            el_dataset_name = el.get("dataset_name")
            el_usecase = el.get("usecase")
            if usecase is not None and el_usecase != usecase:
                continue
            if isinstance(el_model_name, (list, tuple)):
                model_names = el_model_name
            else:
                model_names = [el_model_name]
            if isinstance(el_dataset_name, (list, tuple)):
                dataset_names = el_dataset_name
            else:
                dataset_names = [el_dataset_name]

            model_dataset_pairs = list(itertools.product(model_names, dataset_names))

            for cur_model_name, cur_dataset_name in model_dataset_pairs:
                for test_stage in self._get_list_of_test_stages():
                    test_parameters = deepcopy(el)
                    test_parameters["test_stage"] = test_stage
                    test_parameters["model_name"] = cur_model_name
                    test_parameters["dataset_name"] = cur_dataset_name
                    self._fill_test_parameters_default_values(test_parameters)
                    argvalues.append((test_parameters,))
                    ids.append(self._generate_test_id(test_parameters))

        return argnames, argvalues, ids

    def get_test_case(self, test_parameters, params_factories_for_test_actions):
        """
        The method returns an instance of test case class for the current test parameters.
        If the main test parameters (pointed by the field test_parameters_defining_test_case_behavior)
        are the same as in the previous test, the instance of the test case is re-used from the previous test.
        It allows re-using results of previous tests stages in the next test stages.

        If a new instance of test case should be created, the method creates a class pointed
        by the field test_case_class. The parameter params_factories_for_test_actions is passed
        to the test case constructor
        (note that params_factories_for_test_actions are factories, not structs, to
        create the parameters only when this is required)
        """
        params_defining_cache = {k: test_parameters[k] for k in self.test_parameters_defining_test_case_behavior}
        if not self._cache.has_same_params(params_defining_cache):
            logger.info("OTXTestHelper: parameters were changed -- updating cache")
            logger.info(f"OTXTestHelper: before creating test case (class {self.test_case_class})")
            test_case = self.test_case_class(params_factories_for_test_actions)
            logger.info(f"OTXTestHelper: after creating test case (class {self.test_case_class})")
            self._cache.set(params_defining_cache, test_case)
        else:
            logger.info("OTXTestHelper: parameters were not changed -- cache is kept")

        return self._cache.get()
